#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
from .console import EndlessConsole, Console
import logging
import time
from . import config
from .check import Check
from .test import Test
from github import Github
from zipfile import ZipFile, ZIP_DEFLATED
from tempfile import NamedTemporaryFile
import json
import copy
import hashlib
import requests

class Package():
    """
    Helper for package generation. Helps to create:
     * Cleep .deb package
     * application .zip package
    """

    FRONTEND_DIR = 'frontend/'
    BACKEND_DIR = 'backend/'
    SCRIPTS_DIR = 'scripts/'
    TESTS_DIR = 'tests/'

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.__endless_command_running = False
        self.__endless_command_return_code = 0

    def __compute_sha256(self, fullpath):
        """
        Compute sha256 for specified file

        Args:
            fullpath (string): fullpath

        Returns:
            string: sha256 string
        """
        sha256_hash = hashlib.sha256()
        with open(fullpath,'rb') as f:
            for byte_block in iter(lambda: f.read(4096),b''):
                sha256_hash.update(byte_block)

        return sha256_hash.hexdigest()

    def __console_callback(self, stdout, stderr):
        self.logger.info((stdout if stdout is not None else '') + (stderr if stderr is not None else ''))

    def __console_end_callback(self, return_code, killed):
        self.__endless_command_return_code = return_code
        self.__endless_command_running = False

    def build_cleep(self):
        """
        Build cleep debian package
        """
        self.logger.info('Building Cleep package...')
        cmd = """
SENTRY_DSN=`printenv SENTRY_DSN`

if [ -z "$SENTRY_DSN" ]; then
    echo 
    echo "ERROR: sentry DSN not defined, please set an environment variable called SENTRY_DSN with valid data"
    echo
    exit 1
fi

# Check command result
# $1: command result (usually $?)
# $2: awaited command result
# $3: error message
checkResult() {
    if [ $1 -ne $2 ]
    then
        msg=$3
        if [[ -z "$1" ]]; then
            msg="see output log"
        fi
        echo -e "${RED}Error occured: $msg.${NOCOLOR}"
        exit 1
    fi
}

# clean all files
clean() {
    echo `pwd`
    rm -rf build
    rm -rf .pybuild
    rm -rf debian/cleep
    rm -rf debian/*debhelper*
    rm -rf ../cleep_*_armhf.*
    rm -rf tmp
}

# jump in cleep root directory
cd "%s"

# clean previous build
clean

# check python version
VERSION=`head -n 1 debian/changelog | awk '{ gsub("[\(\)]","",$2); print $2 }'`
PYTHON_VERSION=`cat cleep/__init__.py | grep $VERSION | wc -l`
if [ "$PYTHON_VERSION" -ne "1" ]
then
    echo
    echo "ERROR: python version is not the same than debian version, please update cleep/__init__.py __version__ to $VERSION"
    echo
    exit 1
fi

# generate /etc/default/cleep.conf
mkdir tmp
touch tmp/cleep.conf
echo "SENTRY_DSN=$SENTRY_DSN" >> tmp/cleep.conf

# build Cleep application
debuild -us -uc
checkResult $? 0 "Failed to build deb package"

# clean python stuff
rm -rf cleep.egg-info
rm -rf pycleep.egg-info/
rm -rf tmp/

# jump in build output
cd ".."

# collect variables
DEB=`ls -A1 cleep* | grep \.deb`
# ARCHIVE=cleep_$VERSION.deb
SHA256="cleep_"$VERSION"_armhf.sha256"
# PREINST=cleep/scripts/preinst.sh
# POSTINST=cleep/scripts/postinst.sh

# build zip archive
# rm -f cleep_*.deb
# rm -f cleep_*.sha256
# cp -a $DEB cleep.deb
# cp -a $PREINST .
# cp -a $POSTINST .
# zip $ARCHIVE cleep.deb `basename $PREINST` `basename $POSTINST`
# rm -f `basename $PREINST`
# rm -f `basename $POSTINST`
# rm -f cleep.deb
sha256sum $DEB > $SHA256
        """ % (config.REPO_DIR)
        self.__endless_command_running = True
        c = EndlessConsole(cmd, self.__console_callback, self.__console_end_callback)
        c.start()

        while self.__endless_command_running:
            time.sleep(0.25)

        self.logger.debug('Return code: %s' % self.__endless_command_return_code)
        if self.__endless_command_return_code != 0:
            return False

        return True

    def __delete_github_tag(self, tag_name, token):
        """
        Delete specified tag name (this feature is not available on PyGithub API)

        Args:
            tag_name (string): tag name to delete
            token (string): github authorization token
        """
        # curl request
        # curl -X DELETE -H "Authorization: token <token>" -H "User-Agent: PyGithub/Python" https://api.github.com/repos/tangb/cleep/git/refs/tags/<tag_name>
        try:
            headers = {
                'Authorization': 'token %s' % token,
                'User-Agent': 'PyGithub/Python',
            }
            resp = requests.delete(
                'https://api.github.com/repos/%s/%s/git/refs/tags/%s' % (config.GITHUB_ORG, config.GITHUB_REPO, tag_name),
                headers=headers,
            )
            if resp.status_code >= 200 and resp.status_code < 300:
                return True
            else:
                self.logger.debug('Response data: %s' % resp.data.decode('utf-8'))
                self.logger.error('Unable to delete tag [status=%s]' % resp.status_code)
                return False
        except:
            self.logger.exception('Error deleting tag "%s"' % tag_name)
            return False

    def publish_cleep(self, version):
        """
        Publish cleep version on github
        """
        token = os.environ['GITHUB_ACCESS_TOKEN']
        github = Github(token)
        repo = github.get_repo('%s/%s' % (config.GITHUB_ORG, config.GITHUB_REPO))

        # check build existence
        archive = os.path.abspath(os.path.join(config.REPO_DIR, '..', 'cleep_%s_armhf.deb' % version))
        sha256 = os.path.abspath(os.path.join(config.REPO_DIR, '..', 'cleep_%s_armhf.sha256' % version))
        changes = os.path.abspath(os.path.join(config.REPO_DIR, '..', 'cleep_%s_armhf.changes' % version))
        if not os.path.exists(archive):
            self.logger.error('Archive file "%s" does not exist' % archive)
        if not os.path.exists(sha256):
            self.logger.error('Checksum file "%s" does not exist' % sha256)
        if not os.path.exists(changes):
            self.logger.error('Changes file "%s" does not exist' % changes)

        # get changelog
        cmd = 'sed -n "/cleep (%(version)s)/,/Checksums-Sha1:/{/cleep (%(version)s)/b;/Checksums-Sha1:/b;p}" %(changes)s | tail -n +2' % {'version': version, 'changes': changes}
        self.logger.debug('Cmd = %s' % cmd)
        c = Console()
        result = c.command(cmd)
        if result['error']:
            self.logger.error('Unable to read changelog')
        changelog = '\n'.join([line.strip() for line in result['stdout']])
        self.logger.debug('Changelog:\n%s' % changelog)

        # search existing release
        release_found = None
        releases = repo.get_releases()
        for release in releases:
            self.logger.debug('Release "%s"' % release.title)
            if release.title==version:
                self.logger.debug(' -> Release found')
                release_found = release
                break

        if release_found and (release_found.prerelease or release_found.draft):
            # due to github limitation (bug or limitation?), draft assets are not downloadable
            # so we create prerelease version instead of draft and delete it before creating it
            # again when pushing version
            self.logger.info('Deleting existing release "%s"...' % release_found.tag_name)
            try:
                release_found.delete_release()
            except:
                self.logger.exception('Error deleting existing release:')
                return False

            # delete tag
            if not self.__delete_github_tag(release_found.tag_name, token):
                return False
        
        # create release
        self.logger.info('Creating new release "%s"...' % version)
        try:
            commits = repo.get_commits()
            release_found = repo.create_git_release(
                tag='v%s' % version,
                name=version,
                message=changelog,
                draft=False,
                prerelease=True,
            )
        except:
            self.logger.exception('Error occured creating new release:')
            return False

        # upload assets
        try:
            self.logger.info('Uploading asset "%s"...' % archive)
            release_found.upload_asset(archive)
            self.logger.info('Uploading asset "%s"...' % sha256)
            release_found.upload_asset(sha256)
        except:
            self.logger.exception('Error uploading assets:')
            return False

        return True

        #TODO code for draft release removed for problem downloading draft assets
        #if not release_found:
        #    #create release
        #    self.logger.info('Creating new release "%s"...' % version)
        #    try:
        #        commits = repo.get_commits()
        #        release_found = repo.create_git_release(
        #            tag='v%s' % version,
        #            name=version,
        #            message=changelog,
        #            draft=True,
        #            prerelease=False,
        #        )
        #    except:
        #        self.logger.exception('Error occured creating new release:')
        #        return False

        #else:
        #    #only update draft release !
        #    if not release_found.draft:
        #        self.logger.error('Existing release "%s" is not draft. Please create new release' % version)

        #    #update release data
        #    try:
        #        self.logger.info('Updating existing release "%s"...' % version)
        #        release_found.update_release(
        #            name=version,
        #            message=changelog,
        #            draft=True,
        #            prerelease=False,
        #        )

        #        #delete all assets
        #        self.logger.info('Deleting all existing release assets...')
        #        for asset in release_found.get_assets():
        #            asset.delete_asset()
        #    except:
        #        self.logger.exception('Error occured creating new release:')
        #        return False

    def build_module(self, module_name, ci=False):
        """
        Build module package

        Args:
            module_name (string): module name
            ci (bool): flag for CI (enable tests execution)

        Returns:
            dict: package informations::

            {
                package (string): package fullpath
                sha256 (string): package sha256
                confidence (float): code confidence indicator (based on unit test results)
                quality (float): code quality indicator (based on linter result)
            }

        """
        module_archive = None

        # collect data
        check = Check()
        data_backend = check.check_backend(module_name)
        if len(data_backend['errors']) > 0:
            raise Exception('Error in backend. Fix it before packaging application: %s' % data_backend['errors'])
        data_frontend = check.check_frontend(module_name)
        if len(data_frontend['errors']) > 0:
            raise Exception('Error in frontend. Fix it before packaging application: %s' % data_frontend['errors'])
        data_scripts = check.check_scripts(module_name)
        if len(data_scripts['errors']) > 0:
            raise Exception('Error in scripts. Fix it before packaging application: %s' % data_scripts['errors'])
        data_tests = check.check_tests(module_name)
        if len(data_tests['errors']) > 0:
            raise Exception('Error in tests. Fix it before packaging application: %s' % data_tests['errors'])
        data_changelog = check.check_changelog(module_name)
        logging.debug('Changelog: %s' % data_changelog)
        if data_changelog['version'] != data_backend['metadata']['version']:
            raise Exception('Changelog does not seems to have been updated (no version "%s" found)' % data_backend['metadata']['version'])
        if data_changelog['unreleased']:
            raise Exception('Changelog has UNRELEASED flag enabled, please remove it before publishing')
        data_code_quality = check.check_code_quality(module_name)
        if len(data_code_quality['errors']) > 0:
            raise Exception('Error in code quality. Fix it before packaging application: %s' % data_code_quality['errors'])
        if data_code_quality['score'] < 7.0:
            raise Exception('Code quality for app "%s" is too low to be packaged (%s). Please improve it to be greater than 7.0' % (module_name, data_code_quality['score']))

        data_test = {
            'score': 0.0
        }
        if ci:
            # execute tests
            test = Test()
            if not test.module_test(module_name):
                raise Exception('Test execution failed for "%s". Please fix it.' % module_name)
            data_test = test.module_test_coverage(module_name, as_json=True)

        # build module description file (module.json)
        fdesc = NamedTemporaryFile(delete=False, encoding='utf-8', mode='w')
        module_json = fdesc.name
        metadata = copy.deepcopy(data_backend['metadata'])
        metadata['icon'] = data_frontend['icon']
        metadata['quality'] = data_code_quality['score']
        metadata['confidence'] = data_test['score']
        metadata['changelog'] = data_changelog['changelog']
        fdesc.write(str(json.dumps(metadata, indent=4, ensure_ascii=False, sort_keys=True)))
        fdesc.close()

        # build zip archive
        fdesc = NamedTemporaryFile(delete=False)
        module_archive = fdesc.name
        self.logger.debug('Archive filepath: %s' % module_archive)
        archive = ZipFile(fdesc, 'w', ZIP_DEFLATED)

        # add frontend files
        for a_file in data_frontend['files']:
            archive.write(a_file['fullpath'], os.path.join(self.FRONTEND_DIR, 'js', 'modules', a_file['path']))

        # add backend files
        path_in = data_backend['files']['module']['fullpath']
        path_out = os.path.join(self.BACKEND_DIR, 'modules', data_backend['files']['module']['path'])
        archive.write(path_in, path_out)
        for a_file in data_backend['files']['events']:
            path_in = a_file['fullpath']
            path_out = os.path.join(self.BACKEND_DIR, 'modules', a_file['path'])
            archive.write(path_in, path_out)
        for a_file in data_backend['files']['formatters']:
            path_in = a_file['fullpath']
            path_out = os.path.join(self.BACKEND_DIR, 'modules', a_file['path'])
            archive.write(path_in, path_out)
        for a_file in data_backend['files']['drivers']:
            path_in = a_file['fullpath']
            path_out = os.path.join(self.BACKEND_DIR, 'modules', a_file['path'])
            archive.write(path_in, path_out)
        for a_file in data_backend['files']['misc']:
            path_in = a_file['fullpath']
            path_out = os.path.join(self.BACKEND_DIR, 'modules', a_file['path'])
            archive.write(path_in, path_out)

        # add tests
        for a_file in data_tests['files']:
            path_in = a_file['fullpath']
            path_out = a_file['path']
            archive.write(path_in, path_out)

        # add scripts
        for a_file in data_scripts['files']:
            path_in = a_file['fullpath']
            path_out = a_file['path']
            archive.write(path_in, path_out)

        # add module.json
        archive.write(module_json, 'module.json')

        # close archive
        archive.close()
        archive_path = os.path.join(os.path.dirname(module_archive), 'cleepapp_%s_v%s.zip' % (module_name, metadata['version']))
        os.rename(module_archive, archive_path)

        # clean some stuff
        if os.path.exists(module_json):
            os.remove(module_json)

        self.logger.debug('Package for app "%s" has been built into "%s"' % (module_name, archive_path))

        return {
            'package': archive_path,
            'sha256': self.__compute_sha256(archive_path),
            'quality': metadata['quality'],
            'confidence': metadata['confidence'],
        }




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
from .console import Console
import logging
from . import config
import shutil
import re

class Module():
    """
    Module class
    """

    GIT_IGNORE = """**/*.pyc
**/.coverage
**/.coveralls.yml
docs/*.zip
docs/*.txt
docs/*.xml
docs/source
docs/_build
.vscode
**/*.swp
"""
    CHANGELOG = """# Changelog

## [UNRELEASED] - YYYY-MM-DD
### Added
- First release

    """
    DESC_SKEL = """{
    \\"icon\\": \\"help-circle-outline\\",
    \\"global\\": {
        \\"js\\": [\\"%(MODULE_NAME)s.service.js\\"],
        \\"html\\": [],
        \\"css\\": []
    },
    \\"config\\": {
        \\"js\\": [\\"%(MODULE_NAME)s.config.js\\"],
        \\"html\\": [\\"%(MODULE_NAME)s.config.html\\"],
        \\"css\\": []
    }
}
    """
    ANGULAR_SERVICE_SKEL = """/**
 * %(MODULE_NAME_CAPITALIZED)s service.
 * Handle %(MODULE_NAME)s application requests.
 * Service is the place to store your application content (it is a singleton) and
 * to provide your application functions.
 */
angular
.module('Cleep')
.service('%(MODULE_NAME)sService', ['\$rootScope', 'rpcService',
function(\$rootScope, rpcService) {
    var self = this;

    /**
     * Call backend execute_command command
     */
    self.executeCommand = function() {
        return rpcService.sendCommand('execute_command', '%(MODULE_NAME)s');
    };

    /**
     * Call backend configure_value command
     */
    self.configureValue = function(value) {
        return rpcService.sendCommand('configure_value', '%(MODULE_NAME)s', {
            'value': value,
        });
    };

    self.addFile = function(file) {
        // it uses provided rpcservice upload function
        // add_file refers to your python function that will be called
        return rpcService.upload('add_file', '%(MODULE_NAME)s', file);
    };

    /**
     * Catch an.app.event event
     */
    /*\$rootScope.\$on('an.app.event', function(event, uuid, params) {
        console.log('Event an.app.event received with parameters:', parameters);
    });*/
}]);"""
    ANGULAR_CONTROLLER_SKEL = """/**
 * %(MODULE_NAME_CAPITALIZED)s config component
 * Handle %(MODULE_NAME)s application configuration
 * If your application doesn't need configuration page, delete this file and its references into desc.json
 */
angular
.module('Cleep')
.directive('%(MODULE_NAME)sConfigComponent', ['\$rootScope', 'cleepService', 'toastService', '%(MODULE_NAME)sService',
function(\$rootScope, cleepService, toastService, %(MODULE_NAME)sService) {

    var %(MODULE_NAME)sConfigController = function(\$scope) {
        var self = this;
        // use this config variable in all your app (controller and view) to always have up-to-date config values
        self.config = {};
        self.message = 'message';
        self.checkbox = true;
        self.slider = 25;
        self.addFile = null;

        /**
         * Init component
         */
        self.\$onInit = function() {
            // get module config
            cleepService.getModuleConfig('%(MODULE_NAME)s');
        };

        /**
         * Keep app configuration in sync
         */
        \$rootScope.\$watch(function() {
            return cleepService.modules['%(MODULE_NAME)s'].config;
        }, function(newVal, oldVal) {
            if(newVal && Object.keys(newVal).length) {
                Object.assign(self.config, newVal);
            }
        });

        self.setMessage = function() {
            %(MODULE_NAME)sService.configureValue(self.message)
                .then(function(response) {
                    console.log('Response:', response);
                    if (response.error) {
                        toastService.error('Command failed: ' + response.message);
                        return;
                    }
                    toastService.success('Command succeed');
                })
                .finally(function() {
                    cleepService.reloadModuleConfig('%(MODULE_NAME)s');
                });
        };

        self.executeCommand = function() {
            %(MODULE_NAME)sService.executeCommand()
                .finally(function() {
                    cleepService.reloadModuleConfig('%(MODULE_NAME)s');
                });
        };

        self.updateCheckbox = function() {
            %(MODULE_NAME)sService.configureValue(self.checkbox)
                .finally(function() {
                    cleepService.reloadModuleConfig('%(MODULE_NAME)s');
                });
        };

        self.setSlider = function() {
            %(MODULE_NAME)sService.configureValue(self.slider)
                .finally(function() {
                    cleepService.reloadModuleConfig('%(MODULE_NAME)s');
                });
        };

        /**
         * Handle file upload. Please don't forget to declare self.addFile (see above)
         */
        \$scope.\$watch(function() {
            return self.addFile;
        }, function(file) {
            if (!file) {
                return;
            }

            %(MODULE_NAME)sService.addFile(file)
                .then((response) => {
                    if (!response.error) {
                        toastService.success('File added');
                        // surely need to refresh something ;-)
                    }
                });
        });
    };

    return {
        templateUrl: '%(MODULE_NAME)s.config.html',
        replace: true,
        scope: true,
        controller: %(MODULE_NAME)sConfigController,
        controllerAs: '%(MODULE_NAME)sCtl',
    };
}]);"""
    ANGULAR_CONTROLLER_TEMPLATE_SKEL = """<div layout=\\"column\\" layout-padding ng-cloak>

    <md-list ng-cloak>

        <!-- A configuration section. It is a good way to make separation between different kind of parameters -->
        <md-subheader class=\\"md-no-sticky\\">A section header</md-subheader>

        <!-- list item with single line description -->
        <md-list-item>
            <!-- generic cleep icon, use it as default line icon -->
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <p>One line description</p>
        </md-list-item>

        <!-- list item with double line description -->
        <md-list-item class=\\"md-2-line\\">
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <div class=\\"md-list-item-text\\">
                <h3>Two line description</h3>
                <p>Sub line uses smallest font</p>
            </div>
        </md-list-item>

        <!-- list item with button -->
        <md-list-item>
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <p>Button</p>
            <md-button ng-click=\\"%(MODULE_NAME)sCtl.executeCommand()\\" class=\\"md-raised md-primary md-secondary\\">
                <md-icon md-svg-icon=\\"cog\\"></md-icon>
                Button label
            </md-button>
        </md-list-item>

        <!-- list item with text input -->
        <md-list-item>
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <p>Text input</p>
            <md-input-container md-no-float class=\\"md-secondary no-margin\\" layout=\\"row\\" layout-align=\\"start center\\" layout-padding>
                <div class=\\"no-md-error\\">
                    <input ng-model=\\"%(MODULE_NAME)sCtl.message\\" placeholder=\\"message\\" class=\\"no-margin\\">
                </div>
                <md-button ng-click=\\"%(MODULE_NAME)sCtl.setMessage()\\" class=\\"md-raised md-primary\\">
                    <md-icon md-svg-icon=\\"pencil\\"></md-icon>
                </md-button>
            </md-input-container>
        </md-list-item>

        <!-- list item with checkbox -->
        <md-list-item>
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <p>Checkbox</p>
            <md-checkbox
                class=\\"md-secondary\\"
                ng-model=\\"%(MODULE_NAME)sCtl.checkbox\\"
                ng-change=\\"%(MODULE_NAME)sCtl.updateCheckbox()\\"
            >
            </md-checkbox>
        </md-list-item>

        <!-- list item with slider -->
        <md-list-item>
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <p>Slider</p>
            <div class=\\"md-secondary\\">
                <md-slider
                    ng-model=\\"%(MODULE_NAME)sCtl.slider\\"
                    ng-model-options=\\"{debounce: 1000}\\"
                    ng-change=\\"%(MODULE_NAME)sCtl.setSlider()\\"
                    min=\\"0\\"
                    max=\\"100\\"
                    md-discrete
                >
                </md-slider>
            </div>
        </md-list-item>

        <!-- list item with multiple actions -->
        <md-list-item>
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <p>Multiple actions on the same line</p>
            <md-input-container md-no-float class=\\"md-secondary no-margin\\" layout=\\"row\\" layout-align=\\"start center\\" layout-padding>
                <md-button class=\\"md-raised md-primary\\" ng-click=\\"\\">
                    <md-icon md-svg-icon=\\"delete\\"></md-icon>
                    <md-tooltip>Delete item</md-tooltip>
                </md-button>
                <md-button class=\\"md-raised md-primary\\" ng-click=\\"\\">
                    <md-icon md-svg-icon=\\"pencil\\"></md-icon>
                    <md-tooltip>Edit item</md-tooltip>
                </md-button>
            </md-input-container>
        </md-list-item>

        <!-- upload file example -->
        <md-list-item>
            <md-icon md-svg-icon=\\"chevron-right\\"></md-icon>
            <p>Upload a file to your application</p>
            <md-input-container md-no-float class=\\"md-secondary no-margin\\">
                <div upload-file selected-file=\\"%(MODULE_NAME)sCtl.addFile\\" label=\\"Add file\\"></div>
            </md-input-container>
        </md-list-item>

    </md-list>

</div>
    """
    PYTHON_MODULE_SKEL = """#!/usr/bin/env python
# -*- coding: utf-8 -*-

from cleep.exception import MissingParameter, InvalidParameter, CommandError
from cleep.core import CleepModule
from cleep.common import CATEGORIES

class %(MODULE_NAME_CAPITALIZED)s(CleepModule):
    \\"\\"\\"
    %(MODULE_NAME_CAPITALIZED)s application
    \\"\\"\\"
    MODULE_AUTHOR = 'TODO'
    MODULE_VERSION = '0.0.0'
    MODULE_DEPS = []
    MODULE_DESCRIPTION = 'TODO'
    MODULE_LONGDESCRIPTION = 'TODO'
    MODULE_TAGS = []
    MODULE_CATEGORY = CATEGORIES.APPLICATION
    MODULE_URLINFO = 'https://www.google.com'
    MODULE_URLHELP = None
    MODULE_URLSITE = None
    MODULE_URLBUGS = None

    MODULE_CONFIG_FILE = '%(MODULE_NAME)s.conf'
    DEFAULT_CONFIG = {}

    def __init__(self, bootstrap, debug_enabled):
        \\"\\"\\"
        Constructor

        Args:
            bootstrap (dict): bootstrap objects
            debug_enabled: debug status
        \\"\\"\\"
        CleepModule.__init__(self, bootstrap, debug_enabled)

    def _configure(self):
        \\"\\"\\"
        Configure module.
        Use this function to configure your variables and local stuff that is not blocking.
        At this time other applications are not started and all your command requests will fail.
        \\"\\"\\"
        pass

    def _on_start(self):
        \\"\\"\\"
        Start module.
        Use this function to start your tasks.
        At this time all applications are started and should respond to your command requests.
        \\"\\"\\"
        pass

    def _on_stop(self):
        \\"\\"\\"
        Stop module
        Use this function to stop your tasks and close your external connections.
        \\"\\"\\"
        pass

    def on_event(self, event):
        \\"\\"\\"
        Event received

        Args:
            event (MessageRequest): event data
        \\"\\"\\"
        # execute here actions when you receive a specific event:
        #  - on time event => cron task
        #  - on alert event => send email or push message
        #  - ...
        pass

    def execute_command(self):
        \\"\\"\\"
        This is a simple command execution

        Returns:
            bool: returns always True
        \\"\\"\\"
        self.logger.info('execute_command called')
        return True

    def configure_value(self, value):
        \\"\\"\\"
        This function is an example of how to pass value from frontend to backend

        Args:
            value (any): a value
        \\"\\"\\"
        self.logger.info('set_value called with param value=%%s' %% value)

    def _unexposed_command(self):
        \\"\\"\\"
        A function that starts with _ if not exposed to outside. But it can be used inherited class
        \\"\\"\\"
        self.logger.info('Unexposed function')

    def __another_unexposed_command(self):
        \\"\\"\\"
        A function that starts with __ if not exposed to outside. This kind of function is also declared as private in Python
        and cannot be used elsewhere than in this class itself.
        \\"\\"\\"
        self.logger.info('Another unexposed function')

    def add_file(self, filepath):
        \\"\\"\\"
        Handle upload from your frontend service
        Parameter filepath gives you path to uploaded file.
        Charge to you to move or delete file after processing it.
        \\"\\"\\"
        self.logger.info('File %%s uploaded', filepath)
"""
    TEST_DEFAULT = """#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
import logging
import sys
sys.path.append('../')
from backend.%(MODULE_NAME)s import %(MODULE_NAME_CAPITALIZED)s
from cleep.exception import InvalidParameter, MissingParameter, CommandError, Unauthorized
from cleep.libs.tests import session
from mock import Mock, patch

# Unit testing is part of a development, it's why Cleep requires to have application code tested to
# guarantee a certain source code quality.
#
# If you new to unit testing, you can find a good introduction here https://realpython.com/python-testing/
# Cleep uses unittest framework for which you can find documentation here https://docs.python.org/3/library/unittest.html
#
# You can launch all your tests manually using this command:
#   python3 -m unittest test_%(MODULE_NAME)s.Test%(MODULE_NAME_CAPITALIZED)s
# or a specific test with command:
#   python3 -m unittest test_%(MODULE_NAME)s.Test%(MODULE_NAME_CAPITALIZED)s.test__on_configure
# You can get tests coverage with command:
#   coverage run --omit="*/lib/python*/*","test_*" --concurrency=thread test_%(MODULE_NAME)s.py; coverage report -m -i
# or you can simply use developer application that allows you to perform all tests and coverage directly from web interface
class Test%(MODULE_NAME_CAPITALIZED)s(unittest.TestCase):

    def setUp(self):
        # Change here logging.DEBUG to logging.FATAL to disable logging during tests writings
        # Note that coverage command does not display logging
        logging.basicConfig(level=logging.DEBUG, format='%%(asctime)s %%(name)s:%%(lineno)d %%(levelname)s : %%(message)s')
        self.session = session.TestSession(self)

    def tearDown(self):
        # clean session
        self.session.clean()

    def init(self, start=True):
        \\"\\"\\"
        Call this function at beginning of every test cases. By default it starts your app, but if you specify start=False,
        the application must be started manually which is useful in some cases like testing _on_configure app function.
        \\"\\"\\"
        # next line instanciates your module, overwriting all useful stuff to isolate your module for tests
        self.module = self.session.setup(%(MODULE_NAME_CAPITALIZED)s)
        if start:
            self.session.start_module(self.module)

    # Write your tests here defining functions starting with \\"test_\\"
    # See official documentation https://docs.python.org/3/library/unittest.html
    # def test__on_configure(self):
    #   self.init(start=False)
    #   # create your mocks...
    #
    #   self.module._on_configure()
    #
    #   # check your mocks

    # Write another test. A test case always starts with \\"test_\\"
    # def test_my_function(self):
    #   self.init()
    #   # create your mocks
    #
    #   result = self.module.my_function()
    #
    #   # checks your mocks or function result

# do not remove code below, otherwise tests won't run
if __name__ == '__main__':
    unittest.main()
    """
    DOCS_CONF_PY = """# -*- coding: utf-8 -*-
import os, sys
sys.path.insert(0, os.path.abspath('../'))

project = ''
copyright = ''
author = ''
version = ''
release = ''
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
]
source_suffix = '.rst'
master_doc = 'index'
language = None
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    '**/*.pyc',
]
add_module_names = False
pygments_style = None
html_static_path = [
    '_static',
]
html_css_files = [
    '_static/cleep.css',
]
html_theme = 'sphinx_rtd_theme'
todo_include_todos = True
html_theme_options = {
    'prev_next_buttons_location': None,
    'style_external_links': True,
    'style_nav_header_background': '#607d8b',
    'collapse_navigation': False,
    'sticky_navigation': True,
    'navigation_depth': 6,
    'includehidden': False,
}

def setup(app):
    app.add_css_file('cleep.css')
    """
    DOCS_INDEX_RST = """%(TITLE)s
%(LINE)s

.. toctree::
   :maxdepth: 3
   :caption: Contents:

   source/modules

Indices and tables
==================

* :ref:\`genindex\`
* :ref:\`modindex\`
* :ref:\`search\`
    """
    DOCS_CLEEP_CSS = """.wy-nav-side {
    background: #90a4ae !important;
}
.wy-side-nav-search {
    background: #607d8b !important;
}
.rst-content dl:not(.docutils) dt {
    color: #d32f2f !important;
    background: #ffcdd2 !important;
    border-top: solid 3px #b71c1c !important;
 }
.rst-content dl:not(.docutils) dl dt {
    border: none !important;
    border-left: solid 3px #d32f2f !important;
    background: #eceff1 !important;
    color: #607d8b !important;
}
a:visited, a:hover {
    color: #d32f2f;
}
a {
    color: #000000;
}
a:visited.icon.icon-home, a.icon.icon-home {
    color: #FFFFFF !important;
}
a:hover.icon.icon-home {
    color: #d32f2f;
}
.wy-menu-vertical header,.wy-menu-vertical p.caption {
    color: #000000;
}
    """

    def __init__(self):
        """
        Constructor
        """
        self.logger = logging.getLogger(self.__class__.__name__)

    def create(self, module_name):
        """
        Create module skeleton

        Args:
            module_name (string): module name
        """
        module_name = re.sub('[^0-9a-zA-Z]+', '', module_name).lstrip('0123456789').lower()
        self.logger.debug('Module name to create: %s' % module_name)
        path = os.path.join(config.MODULES_SRC, module_name)
        self.logger.info('Creating module "%s" in "%s"' % (module_name, path))
        
        if os.path.exists(path):
            self.logger.error('Module "%s" already exists in "%s"' % (module_name, path))
            return False

        title = 'Welcome to %s\'s documentation!' % module_name.capitalize()
        templates = {
            'ANGULAR_SERVICE': self.ANGULAR_SERVICE_SKEL % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize()},
            'ANGULAR_CONTROLLER': self.ANGULAR_CONTROLLER_SKEL % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize()},
            'ANGULAR_CONTROLLER_TEMPLATE': self.ANGULAR_CONTROLLER_TEMPLATE_SKEL % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize()},
            'DESC': self.DESC_SKEL % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize()},
            'PYTHON_MODULE': self.PYTHON_MODULE_SKEL % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize()},
            'TEST_DEFAULT': self.TEST_DEFAULT % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize()},
            'DOCS_CONF_PY': self.DOCS_CONF_PY % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize()},
            'DOCS_INDEX_RST': self.DOCS_INDEX_RST % {'MODULE_NAME': module_name, 'MODULE_NAME_CAPITALIZED': module_name.capitalize(), 'TITLE': title, 'LINE': '='*len(title)},
            'DOCS_CLEEP_CSS': self.DOCS_CLEEP_CSS,
            'CHANGELOG': self.CHANGELOG,
            'GIT_IGNORE': self.GIT_IGNORE,
        }

        c = Console()
        resp = c.command("""
/bin/mkdir -p "%(MODULE_DIR)s/backend"
/usr/bin/touch "%(MODULE_DIR)s/backend/__init__.py"
/bin/echo "%(PYTHON_MODULE)s" > %(MODULE_DIR)s/backend/%(MODULE_NAME)s.py
/bin/mkdir -p "%(MODULE_DIR)s/frontend"
/bin/echo "%(DESC)s" > %(MODULE_DIR)s/frontend/desc.json
/bin/echo "%(ANGULAR_SERVICE)s" > %(MODULE_DIR)s/frontend/%(MODULE_NAME)s.service.js
/bin/echo "%(ANGULAR_CONTROLLER)s" > %(MODULE_DIR)s/frontend/%(MODULE_NAME)s.config.js
/bin/echo "%(ANGULAR_CONTROLLER_TEMPLATE)s" > %(MODULE_DIR)s/frontend/%(MODULE_NAME)s.config.html
/bin/mkdir -p "%(MODULE_DIR)s/tests"
/usr/bin/touch "%(MODULE_DIR)s/tests/__init__.py"
/bin/echo "%(TEST_DEFAULT)s" > %(MODULE_DIR)s/tests/test_%(MODULE_NAME)s.py
/bin/mkdir -p "%(MODULE_DIR)s/docs"
/bin/echo "%(DOCS_CONF_PY)s" > %(MODULE_DIR)s/docs/conf.py
/bin/echo "%(DOCS_INDEX_RST)s" > %(MODULE_DIR)s/docs/index.rst
/bin/mkdir -p "%(MODULE_DIR)s/docs/_static"
/bin/echo "%(DOCS_CLEEP_CSS)s" > %(MODULE_DIR)s/docs/_static/cleep.css
/bin/echo "%(CHANGELOG)s" > %(MODULE_DIR)s/CHANGELOG.md
/bin/echo "%(GIT_IGNORE)s" > %(MODULE_DIR)s/.gitignore
""" % {
            'MODULE_DIR': path,
            'MODULE_NAME': module_name,
            'DESC': templates['DESC'],
            'ANGULAR_SERVICE': templates['ANGULAR_SERVICE'],
            'ANGULAR_CONTROLLER': templates['ANGULAR_CONTROLLER'],
            'ANGULAR_CONTROLLER_TEMPLATE': templates['ANGULAR_CONTROLLER_TEMPLATE'],
            'PYTHON_MODULE': templates['PYTHON_MODULE'],
            'TEST_DEFAULT': templates['TEST_DEFAULT'],
            'DOCS_CONF_PY': templates['DOCS_CONF_PY'],
            'DOCS_INDEX_RST': templates['DOCS_INDEX_RST'],
            'DOCS_CLEEP_CSS': templates['DOCS_CLEEP_CSS'],
            'CHANGELOG': templates['CHANGELOG'],
            'GIT_IGNORE': templates['GIT_IGNORE'],
        }, 10)
        if resp['error'] or resp['killed']:
            self.logger.error('Error occured while creating application: %s' % ('killed' if resp['killed'] else resp['stderr']))
            shutil.rmtree(path)
            return False
        
        self.logger.info('Done')
        return True

    def delete(self, module_name):
        """
        Delete installed files for specified  module

        Args:
            module_name (string): module name
        """
        #build all module paths
        paths = [
            os.path.join(config.MODULES_DST, module_name),
            os.path.join(config.MODULES_HTML_DST, module_name),
            os.path.join(config.MODULES_SCRIPTS_DST, module_name),
        ]
        self.logger.info('Deleting module "%s" in "%s"' % (module_name, paths))
        
        deleted = False
        for path in paths:
            if os.path.exists(path):
                shutil.rmtree(path)
                self.logger.debug('Directory "%s" deleted' % path)
                deleted = True

        if not deleted:
            self.logger.info('Nothing has been deleted')

        return True




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import time
from .console import EndlessConsole, Console
import logging
from . import config
import importlib
from datetime import datetime

class Docs():
    """
    Handle documentation processes
    @see https://samnicholls.net/2016/06/15/how-to-sphinx-readthedocs/
    """

    DOCS_TEMP_PATH = '/tmp/cleep-docs'
    DOCS_EXTRACT_PATH = '/tmp/docs'
    DOCS_ARCHIVE_NAME = 'cleep-core-docs.zip'
    DOCS_COMMIT_MESSAGE = 'Update doc Cleep v%(version)s'

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.__endless_command_running = False
        self.__endless_command_return_code = 0
        self.__module_version = None
        self.__module_author = None

    def __console_callback(self, stdout, stderr):
        self.logger.info((stdout if stdout is not None else '') + (stderr if stderr is not None else ''))

    def __console_end_callback(self, return_code, killed):
        self.__endless_command_running = False
        self.__endless_command_return_code = return_code

    def __get_module_data(self, module_name):
        """
        Return useful module data

        Returns:
            dict of data::

                {
                    version (string): module version
                    author (string): module author
                }

        """
        if self.__module_version:
            return {
                'version': self.__module_version,
                'author': self.__module_author
            }

        try:
            module_ = importlib.import_module('cleep.modules.%s' % (module_name))
            app_filename = getattr(module_, 'APP_FILENAME', module_name)
            del module_
            module_ = importlib.import_module('cleep.modules.%s.%s' % (module_name, app_filename))
            class_name = next((item for item in dir(module_) if item.lower() == app_filename.lower()), None)
            module_class_ = getattr(module_, class_name or '', None)
            self.__module_version = module_class_.MODULE_VERSION
            self.__module_author = module_class_.MODULE_AUTHOR

            return {
                'version': self.__module_version,
                'author': self.__module_author
            }
        except:
            self.logger.exception('Unable to get module infos. Is module valid?')
            return None

    def generate_module_docs(self, module_name, preview=False):
        """
        Generate module documentation

        Args:
            module_name (string): module name
            preview (bool): preview generated documentation as text directly on stdout
        """
        #checking module path
        path = os.path.join(config.MODULES_SRC, module_name, 'docs')
        if not os.path.exists(path):
            self.logger.error('Docs directory for module "%s" does not exist' % (module_name))
            return False

        module_data = self.__get_module_data(module_name)
        self.logger.debug('Module data: %s' % module_data)
        if module_data is None:
            return False

        today = datetime.today()

        self.logger.info('=> Generating documentation...')
        cmd = """
cd "%(DOCS_PATH)s"
/usr/bin/rm -rf "%(BUILD_DIR)s" "%(SOURCE_DIR)s"
/usr/local/bin/sphinx-apidoc -o "%(SOURCE_DIR)s/" "../backend"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
echo "=> Building html documentation..."
/usr/local/bin/sphinx-build -M html "." "%(BUILD_DIR)s" -D project="%(MODULE_NAME_CAPITALIZED)s" -D copyright="%(YEAR)s %(AUTHOR)s" -D author="%(AUTHOR)s" -D version="%(VERSION)s" -D release="%(VERSION)s"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
echo "=> Building singlehtml documentation..."
/usr/local/bin/sphinx-build -M singlehtml "." "%(BUILD_DIR)s" -D project="%(MODULE_NAME_CAPITALIZED)s" -D copyright="%(YEAR)s %(AUTHOR)s" -D author="%(AUTHOR)s" -D version="%(VERSION)s" -D release="%(VERSION)s"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
echo "=> Building xml documentation..."
/usr/local/bin/sphinx-build -M xml "." "%(BUILD_DIR)s" -D project="%(MODULE_NAME_CAPITALIZED)s" -D copyright="%(YEAR)s %(AUTHOR)s" -D author="%(AUTHOR)s" -D version="%(VERSION)s" -D release="%(VERSION)s"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
echo "=> Building text documentation..."
/usr/local/bin/sphinx-build -M text "." "%(BUILD_DIR)s" -D project="%(MODULE_NAME_CAPITALIZED)s" -D copyright="%(YEAR)s %(AUTHOR)s" -D author="%(AUTHOR)s" -D version="%(VERSION)s" -D release="%(VERSION)s"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
echo "=> Packaging html documentation..."
/usr/bin/find "%(BUILD_DIR)s/" -type f -print0 | xargs -0 sed -i "s/backend/%(MODULE_NAME)s/g"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
/usr/bin/find "%(BUILD_DIR)s/" -type f -print0 | xargs -0 sed -i "s/Backend/%(MODULE_NAME_CAPITALIZED)s/g"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
/usr/bin/find "%(BUILD_DIR)s/" -iname \*.* | for f in $(find . -name backend*); do mv $f $(echo "$f" | sed -r 's|backend|%(MODULE_NAME)s|g'); done
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
#/bin/tar -czvf "%(MODULE_NAME)s-docs.tar.gz" "%(BUILD_DIR)s/html" --transform='s/%(BUILD_DIR)s\//\//g' && ARCHIVE=`/usr/bin/realpath "%(MODULE_NAME)s-docs.tar.gz"` && echo "ARCHIVE=$ARCHIVE"
cd "_build"; /usr/bin/zip "../%(MODULE_NAME)s-docs.zip" -r "html"; cd ..
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
/bin/cp -a "%(BUILD_DIR)s/text/source/%(MODULE_NAME)s.txt" "%(MODULE_NAME)s-docs.txt"
/bin/cp -a "%(BUILD_DIR)s/xml/source/%(MODULE_NAME)s.xml" "%(MODULE_NAME)s-docs.xml"
%(DISPLAY_TEXT)s
        """ % {
            'DOCS_PATH': path,
            'SOURCE_DIR': 'source',
            'BUILD_DIR': '_build',
            'MODULE_NAME': module_name,
            'MODULE_NAME_CAPITALIZED': module_name.capitalize(),
            'YEAR': today.year,
            'AUTHOR': module_data['author'],
            'VERSION': module_data['version'],
            'DISPLAY_TEXT': 'echo; echo; echo "========== DOC PREVIEW =========="; echo; cat "%s-docs.txt"' % module_name if preview else '',
        }

        self.logger.debug('Docs cmd: %s' % cmd)
        self.__endless_command_running = True
        c = EndlessConsole(cmd, self.__console_callback, self.__console_end_callback)
        c.start()

        while self.__endless_command_running:
            time.sleep(0.25)

        self.logger.debug('Return code: %s' % self.__endless_command_return_code)
        if self.__endless_command_return_code!=0:
            return False

        return True

    def get_module_docs_archive_path(self, module_name):
        """
        Display module docs archive path if exists
        """
        #checking module path
        docs_path = os.path.join(config.MODULES_SRC, module_name, 'docs')
        if not os.path.exists(docs_path):
            self.logger.error('Docs directory for module "%s" does not exist' % (module_name))
            return False

        zip_path = os.path.join(docs_path, '%s-docs.zip' % module_name)
        if not os.path.exists(zip_path):
            self.logger.error('There is no documentation archive generated for module "%s"' % (module_name))
            return False

        self.logger.info('DOC_ARCHIVE=%s' % zip_path)
        return True

    def generate_core_docs(self, publish=False):
        """
        Generate core documentation
        """
        # checking core path
        path = os.path.join(config.CORE_SRC, '../docs')
        self.logger.debug('Core docs path: %s' % path)
        if not os.path.exists(path):
            self.logger.error('Docs directory for core does not exist')
            return False

        today = datetime.today()

        self.logger.info('=> Generating documentation...')
        cmd = """
cd "%(DOCS_PATH)s"
echo "=> Generating documentation sources..."
/bin/rm -rf "%(BUILD_DIR)s" "%(SOURCE_DIR)s"
/usr/local/bin/sphinx-apidoc -o "%(SOURCE_DIR)s/" "../cleep" "../cleep/tests/**" "../cleep/modules/**"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
/bin/rm -f "%(CORE)s-docs.zip"
echo "=> Building html documentation..."
/usr/local/bin/sphinx-build -M html "." "%(BUILD_DIR)s" -D project="%(PROJECT)s" -D copyright="%(YEAR)s %(AUTHOR)s" -D author="%(AUTHOR)s" -D version="%(VERSION)s" -D release="%(VERSION)s"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
echo "=> Building xml documentation..."
/usr/local/bin/sphinx-build -M xml "." "%(BUILD_DIR)s" -D project="%(PROJECT)s" -D copyright="%(YEAR)s %(AUTHOR)s" -D author="%(AUTHOR)s" -D version="%(VERSION)s" -D release="%(VERSION)s"
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
echo
# echo "=> Building text documentation..."
# /usr/local/bin/sphinx-build -M text "." "%(BUILD_DIR)s" -D project="%(PROJECT)s" -D copyright="%(YEAR)s %(AUTHOR)s" -D author="%(AUTHOR)s" -D version="%(VERSION)s" -D release="%(VERSION)s"
# if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
# echo
echo "=> Packaging html documentation..."
cd "_build"; /usr/bin/zip "../%(CORE)s-docs.zip" -r "html"; cd ..
if [ $? -ne 0 ]; then echo "Error occured"; exit 1; fi
# TODO /bin/cp -a "%(BUILD_DIR)s/xml/source/%(CORE)s.xml" "%(CORE)s-docs.xml"
        """ % {
            'DOCS_PATH': path,
            'SOURCE_DIR': 'source',
            'BUILD_DIR': '_build',
            'CORE': 'cleep-core',
            'PROJECT': config.DOCS_PROJECT_NAME,
            'YEAR': today.year,
            'AUTHOR': config.DOCS_AUTHOR,
            'VERSION': config.CORE_VERSION,
        }

        self.logger.debug('Docs cmd: %s' % cmd)
        self.__endless_command_running = True
        c = EndlessConsole(cmd, self.__console_callback, self.__console_end_callback)
        c.start()

        while self.__endless_command_running:
            time.sleep(0.25)

        self.logger.debug('Return code: %s' % self.__endless_command_return_code)
        if self.__endless_command_return_code!=0:
            return False

        # publish docs
        if publish:
            self.logger.info('=> Publishing documentation for Cleep v%s...', config.CORE_VERSION)
            return self.publish_core_docs()

        return True

    def publish_core_docs(self):
        """
        Publish core docs to github pages
        """
        c = Console()

        # check
        docs_archive_path = os.path.join(config.REPO_DIR, 'docs/', self.DOCS_ARCHIVE_NAME)
        if not os.path.exists(docs_archive_path):
            self.logger.error('Core has no docs archive generated (%s). Please run "cleep-cli coredocs" first' % docs_archive_path)
            return False

        # clone repo
        self.logger.debug('Cloning core repository...')
        repo = 'git@github.com:%s/%s.git' % (config.GITHUB_ORG, config.GITHUB_REPO)
        cmd = 'rm -rf "%s"; git clone -q "%s" "%s"' % (self.DOCS_TEMP_PATH, repo, self.DOCS_TEMP_PATH)
        self.logger.debug('cmd: %s' % cmd)
        resp = c.command(cmd, 60) 
        self.logger.debug('Clone resp: %s' % resp)
        if resp['returncode'] != 0 or resp['killed']:
            self.logger.error('Error occured while cloning repository: %s' % ('killed' if resp['killed'] else resp['stderr']))
            return False
    
        # switch to branch
        self.logger.debug('Switching to %s branch' % config.GITHUB_DOCS_BRANCH)
        cmd = 'cd "%s" && git fetch --all && git checkout -b "%s" "origin/%s"' % (self.DOCS_TEMP_PATH, config.GITHUB_DOCS_BRANCH, config.GITHUB_DOCS_BRANCH)
        self.logger.debug('cmd: %s' % cmd)
        resp = c.command(cmd, 60) 
        self.logger.debug('Switch branch resp: %s' % resp)
        if resp['returncode'] != 0 or resp['killed']:
            self.logger.error('Error occured while switching repository branch: %s' % ('killed' if resp['killed'] else resp['stderr']))
            return False

        # add docs
        self.logger.debug('Updating docs...')
        cmd = 'unzip "%s" -d "%s" && rm -rf "%s" && mv "%s" "%s"' % (
            docs_archive_path,
            self.DOCS_EXTRACT_PATH,
            os.path.join(self.DOCS_TEMP_PATH, 'docs'),
            os.path.join(self.DOCS_EXTRACT_PATH, 'html'),
            os.path.join(self.DOCS_TEMP_PATH, 'docs'),
        )
        self.logger.debug('cmd: %s' % cmd)
        resp = c.command(cmd, 60) 
        self.logger.debug('Unzip resp: %s' % resp)
        if resp['returncode'] != 0 or resp['killed']:
            self.logger.error('Error occured while unzipping core docs archive: %s' % ('killed' if resp['killed'] else resp['stderr']))
            return False

        # commit changes
        self.logger.debug('Commiting changes...')
        cmd = 'cd "%s" && git add . && git commit -m "%s" && git push' % (
            self.DOCS_TEMP_PATH,
            self.DOCS_COMMIT_MESSAGE % { 'version': config.CORE_VERSION },
        )
        self.logger.debug('cmd: %s' % cmd)
        resp = c.command(cmd, 60) 
        self.logger.debug('Commit resp: %s' % resp)
        if resp['returncode'] != 0 or resp['killed']:
            self.logger.error('Error occured while pushing changes: %s' % ('killed' if resp['killed'] else resp['stderr']))
            return False

        return True




#!/usr/bin/env python
# -*- coding: utf-8 -*-
   
import binascii
from passlib.utils import pbkdf2
import base64
import io
import os
import logging
import sys
import subprocess

#from https://elinux.org/RPi_HardwareHistory
RASPBERRY_PI_REVISIONS = {
    u'unknown':{u'date': u'?',        u'model': u'?',                                u'pcbrevision': u'?',   u'ethernet': False, u'wireless': False, u'audio':False, u'gpiopins': 0,  u'memory': '?',              u'notes': u'Unknown model'},
    u'0002':   {u'date': u'Q1 2012',  u'model': u'B',                                u'pcbrevision': u'1.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u''},
    u'0003':   {u'date': u'Q3 2012',  u'model': u'B (ECN0001)',                      u'pcbrevision': u'1.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u'Fuses mod and D14 removed'},
    u'0004':   {u'date': u'Q3 2012',  u'model': u'B',                                u'pcbrevision': u'2.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u'(Mfg by Sony)'},
    u'0005':   {u'date': u'Q4 2012',  u'model': u'B',                                u'pcbrevision': u'2.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u'(Mfg by Qisda)'},
    u'0006':   {u'date': u'Q4 2012',  u'model': u'B',                                u'pcbrevision': u'2.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u'(Mfg by Egoman)'},
    u'0007':   {u'date': u'Q1 2013',  u'model': u'A',                                u'pcbrevision': u'2.0', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u'(Mfg by Egoman)'},
    u'0008':   {u'date': u'Q1 2013',  u'model': u'A',                                u'pcbrevision': u'2.0', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u'(Mfg by Sony)'},
    u'0009':   {u'date': u'Q1 2013',  u'model': u'A',                                u'pcbrevision': u'2.0', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'256 MB',        u'notes': u'(Mfg by Qisda)'},
    u'000d':   {u'date': u'Q4 2012',  u'model': u'B',                                u'pcbrevision': u'2.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'512 MB',        u'notes': u'(Mfg by Egoman)'},
    u'000e':   {u'date': u'Q4 2012',  u'model': u'B',                                u'pcbrevision': u'2.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'000f':   {u'date': u'Q4 2012',  u'model': u'B',                                u'pcbrevision': u'2.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 26, u'memory': u'512 MB',        u'notes': u'(Mfg by Qisda)'},
    u'0010':   {u'date': u'Q3 2014',  u'model': u'B+',                               u'pcbrevision': u'1.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'0011':   {u'date': u'Q2 2014',  u'model': u'Compute Module 1',                 u'pcbrevision': u'1.0', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 0,  u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'0012':   {u'date': u'Q4 2014',  u'model': u'A+',                               u'pcbrevision': u'1.1', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'256 MB',        u'notes': u'(Mfg by Sony)'},
    u'0013':   {u'date': u'Q1 2015',  u'model': u'B+',                               u'pcbrevision': u'1.2', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Embest)'},
    u'0014':   {u'date': u'Q2 2014',  u'model': u'Compute Module 1',                 u'pcbrevision': u'1.0', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 0 , u'memory': u'512 MB',        u'notes': u'(Mfg by Embest)'},
    u'0015':   {u'date': u'?',        u'model': u'A+',                               u'pcbrevision': u'1.1', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'256 MB/512 MB', u'notes': u'(Mfg by Embest)'},
    u'a01040': {u'date': u'Unknown',  u'model': u'2 Model B',                        u'pcbrevision': u'1.0', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Sony)'},
    u'a01041': {u'date': u'Q1 2015',  u'model': u'2 Model B',                        u'pcbrevision': u'1.1', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Sony)'},
    u'a21041': {u'date': u'Q1 2015',  u'model': u'2 Model B',                        u'pcbrevision': u'1.1', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Embest)'},
    u'a22042': {u'date': u'Q3 2016',  u'model': u'2 Model B (with BCM2837)',         u'pcbrevision': u'1.2', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Embest)'},
    u'900021': {u'date': u'Q3 2016',  u'model': u'A+',                               u'pcbrevision': u'1.1', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'900032': {u'date': u'Q2 2016?', u'model': u'B+',                               u'pcbrevision': u'1.2', u'ethernet': True,  u'wireless': False, u'audio': True, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'900092': {u'date': u'Q4 2015',  u'model': u'Zero',                             u'pcbrevision': u'1.2', u'ethernet': False, u'wireless': False, u'audio': False, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'900093': {u'date': u'Q2 2016',  u'model': u'Zero',                             u'pcbrevision': u'1.3', u'ethernet': False, u'wireless': False, u'audio': False, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'920093': {u'date': u'Q4 2016?', u'model': u'Zero',                             u'pcbrevision': u'1.3', u'ethernet': False, u'wireless': False, u'audio': False, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Embest)'},
    u'9000c1': {u'date': u'Q1 2017',  u'model': u'Zero W',                           u'pcbrevision': u'1.1', u'ethernet': False, u'wireless': True,  u'audio': False, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'a02082': {u'date': u'Q1 2016',  u'model': u'3 Model B',                        u'pcbrevision': u'1.2', u'ethernet': True,  u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Sony)'},
    u'a020a0': {u'date': u'Q1 2017',  u'model': u'Compute Module 3 (and CM3 Lite)',  u'pcbrevision': u'1.0', u'ethernet': False, u'wireless': False, u'audio': True, u'gpiopins': 0,  u'memory': u'1 GB',          u'notes': u'(Mfg by Sony)'},
    u'a22082': {u'date': u'Q1 2016',  u'model': u'3 Model B',                        u'pcbrevision': u'1.2', u'ethernet': True,  u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Embest)'},
    u'a32082': {u'date': u'Q4 2016',  u'model': u'3 Model B',                        u'pcbrevision': u'1.2', u'ethernet': True,  u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Sony Japan)'},
    u'a020d3': {u'date': u'Q1 2018',  u'model': u'3 Model B+',                       u'pcbrevision': u'1.3', u'ethernet': True,  u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Sony)'},
    u'9020e0': {u'date': u'Q4 2018',  u'model': u'3 Model A+',                       u'pcbrevision': u'1.0', u'ethernet': False, u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'512 MB',        u'notes': u'(Mfg by Sony)'},
    u'a03111': {u'date': u'Q2 2019',  u'model': u'4 Model B',                        u'pcbrevision': u'1.1', u'ethernet': True,  u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'1 GB',          u'notes': u'(Mfg by Sony)'},
    u'b03111': {u'date': u'Q2 2019',  u'model': u'4 Model B',                        u'pcbrevision': u'1.1', u'ethernet': True,  u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'2 GB',          u'notes': u'(Mfg by Sony)'},
    u'c03111': {u'date': u'Q2 2019',  u'model': u'4 Model B',                        u'pcbrevision': u'1.1', u'ethernet': True,  u'wireless': True,  u'audio': True, u'gpiopins': 40, u'memory': u'4 GB',          u'notes': u'(Mfg by Sony)'},
}

def raspberry_pi_infos():
    """
    Returns infos about current raspberry pi board

    Notes:
        https://elinux.org/RPi_HardwareHistory#Board_Revision_History

    Returns:
        dict: raspberry pi board infos::

            {
                date (string): release date
                model (string): raspberry pi model
                pcbrevision (string): PCB revision
                ethernet (bool): True if ethernet is natively available on board
                wireless (bool): True if wifi is natively available on board,
                audio (bool): True if audio is natively available on board
                gpiopins (int): number of pins available on board
                memory (string): memory amount
                notes (string): notes on board
                revision (string): raspberry pi revision
            }

    """
    cmd = u'/usr/bin/awk \'/^Revision/ {sub("^1000", "", $3); print $3}\' /proc/cpuinfo'
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    revision = p.communicate()[0].decode('utf-8').replace(u'\n', u'')
    logging.trace('Raspberrypi revision=%s' % revision)
    infos = RASPBERRY_PI_REVISIONS[revision] if revision and revision in RASPBERRY_PI_REVISIONS else RASPBERRY_PI_REVISIONS[u'unknown']
    infos[u'revision'] = revision

    return infos

TRACE = logging.DEBUG - 5
        
def install_trace_logging_level():
    """ 
    Install custom log level TRACE for library debugging principaly
    Credits https://gist.github.com/numberoverzero/f803ebf29a0677b6980a5a733a10ca71
    """
    level = logging.TRACE = TRACE 

    def log_logger(self, message, *args, **kwargs):
        if self.isEnabledFor(level):
            self._log(level, message, args, **kwargs)
    logging.getLoggerClass().trace = log_logger

    def log_root(msg, *args, **kwargs):
        logging.log(level, msg, *args, **kwargs)
    logging.addLevelName(level, "TRACE")
    logging.trace = log_root

def install_unhandled_exception_handler(crash_report): # pragma: no cover (can test it)
    """
    Overwrite default exception handler to log errors
    @see https://stackoverflow.com/a/16993115

    Args:
        crash_report (CrashReport): crash report instance
    """
    def handle_exception(exc_type, exc_value, exc_traceback):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        if issubclass(exc_type, KeyboardInterrupt):
            return
        if crash_report:
            crash_report.report_exception()
        logging.error('Uncaught exception', exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = handle_exception

DBM_TO_PERCENT = {-1:100, -2:100, -3:100, -4:100, -5:100, -6:100, -7:100, -8:100, -9:100, -10:100, -11:100, -12:100, -13:100, -14:100, -15:100, -16:100, -17:100, -18:100, -19:100, -20:100, -21:99, -22:99, -23:99, -24:98, -25:98, -26:98, -27:97, -28:97, -29:96, -30:96, -31:95, -32:95, -33:94, -34:93, -35:93, -36:92, -37:91, -38:90, -39:90, -40:89, -41:88, -42:87, -43:86, -44:85, -45:84, -46:83, -47:82, -48:81, -49:80, -50:79, -51:78, -52:76, -53:75, -54:74, -55:73, -56:71, -57:70, -58:69, -59:67, -60:66, -61:64, -62:63, -63:61, -64:60, -65:58, -66:56, -67:55, -68:53, -69:51, -70:50, -71:48, -72:46, -73:44, -74:42, -75:40, -76:38, -77:36, -78:34, -79:32, -80:30, -81:28, -82:26, -83:24, -84:22, -85:20, -86:17, -87:15, -88:13, -89:10, -90:8, -91:6, -92:3, -93:1, -94:1, -95:1, -96:1, -97:1, -98:1, -99:1, -100:1}

def dbm_to_percent(dbm):
    """
    Convert dbm signal level to percentage

    Notes:
        Article here https://www.adriangranados.com/blog/dbm-to-percent-conversion

    Args:
        dbm (int): dbm value

    Returns:
        int: percentage value
    """
    if dbm in DBM_TO_PERCENT.keys():
        return DBM_TO_PERCENT[dbm]

    return 0

def wpa_passphrase(ssid, password):
    """
    Python implementation of wpa_passphrase linux utility
    It generates wpa_passphrase for wifi network connection

    Notes:
        Copied from https://github.com/julianofischer/python-wpa-psk-rawkey-gen/blob/master/rawkey-generator.py

    Args:
        ssid (string): network ssid
        password (string): password

    Returns:
        string: generated psk
    """
    psk = pbkdf2.pbkdf2(str.encode(password), str.encode(ssid), 4096, 32)
    return binascii.hexlify(psk).decode("utf-8")

def file_to_base64(path):
    """
    Convert specified file to base64 string

    Args:
        path (string): path to file

    Returns:
        string: base64 encoded file content

    Raises:
        Exception of all kind if something wrong occured
    """
    with io.open(path, u'rb') as file_to_convert:
        return base64.b64encode(file_to_convert.read()).decode('utf-8')

def hr_uptime(uptime):
    """  
    Human readable uptime (in days/hours/minutes/seconds)

    Notes:
        http://unix.stackexchange.com/a/27014

    Args:
        uptime (int): uptime value

    Returns:
        string: human readable string
    """
    #get values
    days = uptime / 60 / 60 / 24 
    hours = uptime / 60 / 60 % 24 
    minutes = uptime / 60 % 60 

    return u'%dd %dh %dm' % (days, hours, minutes)

def hr_bytes(n):
    """  
    Human readable bytes value

    Notes:
        http://code.activestate.com/recipes/578019

    Args:
        n (int): bytes

    Returns:
        string: human readable bytes value
    """
    symbols = (u'K', u'M', u'G', u'T', u'P', u'E', u'Z', u'Y')
    prefix = {} 

    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10 

    for s in reversed(symbols):
        if n >= prefix[s]:
            value = float(n) / prefix[s]
            return u'%.1f%s' % (value, s)

    return u'%sB' % n

def compare_versions(old_version, new_version):
    """ 
    Compare specified version and return True if new version is strictly greater than old one

    Args:
        old_version (string): old version
        new_version (string): new version

    Returns:
        bool: True if new version available
    """
    #check versions
    old_vers = tuple(map(int, (old_version.split(u'.'))))
    if len(old_vers)!=3:
        raise Exception('Invalid version "%s" format, only 3 digits format allowed' % old_version)

    new_vers = tuple(map(int, (new_version.split(u'.'))))
    if len(new_vers)!=3:
        raise Exception('Invalid version "%s" format, only 3 digits format allowed' % new_version)

    #compare version
    if old_vers<new_vers:
        return True

    return False

def full_split_path(path):
    """
    Split path completely /home/test/test.txt => ['/', 'home', 'test', 'test.py']

    Notes:
        code from https://www.safaribooksonline.com/library/view/python-cookbook/0596001673/ch04s16.html

    Returns:
        list: list of path parts
    """
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])
    
    return list(filter(lambda p: len(p)>0, allparts))

def netmask_to_cidr(netmask):
    """ 
    Convert netmask to cidr format

    Notes:
        code from https://stackoverflow.com/a/43885814

    Args:
        netmask (string): netmask address

    Returns:
        int: cidr value
    """
    return sum([bin(int(x)).count('1') for x in netmask.split('.')])

def cidr_to_netmask(cidr):
    """
    Convert cidr to netmask

    Notes:
        http://www.linuxquestions.org/questions/blog/bittner-195120/cidr-to-netmask-conversion-with-python-convert-short-netmask-to-long-dotted-format-3147/

    Args:
        cidr (int): cidr value

    Returns:
        string: netmask (ie 255.255.255.0)
    """
    mask = ''
    if not isinstance(cidr, int) or cidr<0 or cidr>32: # pragma: no cover
            return None
    
    for t in range(4):
        if cidr > 7:
            mask += '255.'
        else:
            dec = 255 - (2**(8 - cidr) - 1)
            mask += str(dec) + '.' 
        cidr -= 8
        if cidr < 0:
            cidr = 0 
    
    return mask[:-1]




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from importlib.machinery import SourceFileLoader
try:
    import cleep
except:
    class cleep:
        __file__ = None

def get_core_version_from_sources(repo_dir):
    try:
        cleep_sources = SourceFileLoader("cleep_sources", os.path.join(repo_dir, 'cleep', '__init__.py')).load_module()
        return cleep_sources.__version__
    except:
        return None

DEFAULT_MODULES = ['system', 'parameters', 'cleepbus', 'audio', 'network', 'update']
MODULES_REPO_URL = {
    'system': 'https://github.com/CleepDevice/cleepapp-system.git',
    'parameters': 'https://github.com/CleepDevice/cleepapp-parameters.git',
    'cleepbus': 'https://github.com/CleepDevice/cleepapp-cleepbus.git',
    'audio': 'https://github.com/CleepDevice/cleepapp-audio.git',
    'network': 'https://github.com/CleepDevice/cleepapp-network.git',
    'update': 'https://github.com/CleepDevice/cleepapp-update.git',
}

GITHUB_ORG = 'CleepDevice'
GITHUB_REPO = 'cleep'
GITHUB_DOCS_BRANCH = 'docs'

REPO_URL = 'https://github.com/CleepDevice/cleep.git'
REPO_DIR = os.environ.get('REPO_DIR', '/root/cleep-dev')

CORE_SRC = '%s/cleep' % REPO_DIR
CORE_DST = os.path.dirname(cleep.__file__) if cleep.__file__ else None

HTML_SRC = '%s/html' % REPO_DIR
HTML_DST = '/opt/cleep/html'

MODULES_SRC = '%s/modules' % REPO_DIR
MODULES_DST = '%s/modules' % CORE_DST
MODULES_HTML_DST = '%s/js/modules' % HTML_DST
MODULES_SCRIPTS_DST = '/opt/cleep/scripts'

BIN_SRC = '%s/bin' % REPO_DIR
BIN_DST = '/usr/bin'

MEDIA_SRC = '%s/medias' % REPO_DIR
MEDIA_DST = '/opt/cleep'

DOCS_AUTHOR = 'CleepDevice'
DOCS_PROJECT_NAME = 'Cleep core'

CORE_VERSION = get_core_version_from_sources(REPO_DIR)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import time
from .console import EndlessConsole, Console, AdvancedConsole
import logging
from . import config
import importlib
import re
import datetime
import shutil

class Test():
    """
    Handle test processes
    """
    COVERAGE_PATH = '/opt/cleep/.coverage'

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.__endless_command_running = False
        self.__endless_command_return_code = 0
        self.__module_version = None
        self.__stderr = []
        if not os.path.exists(self.COVERAGE_PATH):
            os.makedirs(self.COVERAGE_PATH)

    def __coverage_to_dict(self, stdout):
        """
        Convert coverage results to dict

        Args:
            stdout (string): coverage results
        """
        out = {
            'files': [],
            'score': 0.0,
        }

        c = AdvancedConsole()
        results = c.find_in_string(stdout, r'((?:/.*?)+\.py)\s+(\d+)\s+\d+\s+(\d+)%|^(TOTAL)\s+\d+\s+\d+\s+(\d+)%$')
        for result in results:
            groups = list(filter(None, result[1]))
            if groups[0].startswith('TOTAL'):
                out['score'] = int(groups[1]) / 10
            else:
                out['files'].append({
                    'file': groups[0],
                    'statements': int(groups[1]),
                    'coverage': int(groups[2]),
                })

        return out

    def __reset_stds(self):
        self.__stderr.clear()

    def __quiet_console_callback(self, stdout, stderr):
        if stderr:
            self.__stderr.append(stderr)

    def __console_callback(self, stdout, stderr):
        if stdout and stdout.find('----------------------------------------------------------------------') < 0:
            self.logger.info("  |  " + stdout)
        if stderr and stderr.find('----------------------------------------------------------------------') < 0:
            self.logger.info("  | " + stderr)
        if stderr:
            self.__stderr.append(stderr)

    def __console_end_callback(self, return_code, killed):
        self.__endless_command_running = False
        self.__endless_command_return_code = return_code

    def __get_module_version(self, module_name):
        if self.__module_version:
            return self.__module_version

        try:
            module_ = importlib.import_module('cleep.modules.%s' % (module_name))
            app_filename = getattr(module_, 'APP_FILENAME', module_name)
            del module_
            module_ = importlib.import_module('cleep.modules.%s.%s' % (module_name, app_filename))
            class_name = next((item for item in dir(module_) if item.lower() == app_filename.lower()), None)
            module_class_ = getattr(module_, class_name or '', None)
            self.__module_version = module_class_.MODULE_VERSION
            return self.__module_version
        except:
            self.logger.exception('Unable to get module version. Is module valid?')
            return None

    def __get_coverage_errors_and_failures(self, test_filepath):
        """
        Parse coverage output to find errors and failures count

        Args:
            test_filepath (string): test_filepath

        Returns:
            tuple: errors and failures count::

                (errors, failures)

        """
        exception = ''
        errors = 0
        failures = 0
        notest = False

        # coverage output is stored in stderr
        pattern = r'^FAILED \(((?:failures=(?P<failures>\d+))?|(?:, )?|(?:errors=(?P<errors>\d+))?)*\)$'
        matches = re.finditer(pattern, '\n'.join(self.__stderr), re.MULTILINE)
        for matchNum, match in enumerate(matches, start=1):
            indexes = match.groupdict()
            errors = indexes['errors'] if indexes['errors'] else 0
            failures = indexes['failures'] if indexes['failures'] else 0

        if not os.path.exists(test_filepath):
            notest = 'no test, '
            exception = ''
        else:
            notest = ''
            exception = 'run error, ' if errors == 0 and failures == 0 else ''

        return (errors, failures, exception, notest)

    def __get_coverage_file(self, module_name, module_version):
        """
        Return coverage file

        Args:
            module_name (string): module name
            module_version (string): module version

        Returns:
            string: coverage file
        """
        cov_file = os.path.join(self.COVERAGE_PATH, '%s.%s.cov' % (module_name, module_version))
        self.logger.debug('Coverage file = %s' % cov_file)
        return cov_file

    def __get_module_tests_path(self, module_name):
        """
        Return module tests path

        Args:
            module_name (string): module name

        Returns:
            string: module tests path
        """
        return '%s/%s/tests' % (config.MODULES_SRC, module_name)

    def __coverage_simple_command(self, path, coverage_command, coverage_options='', coverage_file=None, timeout=5.0):
        coverage_file = 'COVERAGE_FILE=%s ' % coverage_file if coverage_file else ''
        
        cmd = """
cd "%(path)s"
%(coverage_file)scoverage %(command)s %(options)s
        """ % {
            'coverage_file': coverage_file,
            'path': path,
            'command': coverage_command,
            'options': coverage_options,
        }
        c = Console()
        res = c.command(cmd, timeout=timeout)
        if res['error'] or res['killed']:
            self.logger.error('Error during command execution (killed: %s): %s' % (res['killed'], res['stderr']))
            return False

        return res

    def module_test(self, module_name, display_coverage=False, copy_to=None):
        """
        Execute module unit tests and display process output on stdout

        Args:
            module_name (string): module name
            display_coverage (bool): display coverage report (default False)

        Returns:
            bool: True if process succeed, False otherwise
        """
        # checking module path
        path = os.path.join(config.MODULES_SRC, module_name, 'tests')
        if not os.path.exists(path):
            self.logger.error('Specified module "%s" does not exist' % (module_name))
            return False

        # create module coverage path
        module_version = self.__get_module_version(module_name)
        if module_version is None:
            return False

        self.logger.info('Running unit tests...')
        coverage_file_path = self.__get_coverage_file(module_name, module_version)
        module_tests_path = self.__get_module_tests_path(module_name)
        cmd = """
cd "%s"
COVERAGE_FILE=%s coverage run --omit="*/lib/python*/*","test_*" --source="../backend" --concurrency=thread -m unittest discover
        """ % (module_tests_path, coverage_file_path)
        self.logger.debug('Test cmd: %s' % cmd)
        self.__endless_command_running = True
        self.__reset_stds()
        c = EndlessConsole(cmd, self.__console_callback, self.__console_end_callback)
        c.start()

        while self.__endless_command_running:
            time.sleep(0.25)

        # display coverage report
        if display_coverage:
            self.logger.debug('Display coverage')
            self.logger.info(self.module_test_coverage(module_name))

        # copy coverage file to standart location (in module tests directory) for coverage report tools
        if copy_to:
            shutil.copyfile(coverage_file_path, os.path.join(copy_to, '.coverage'))

        self.logger.debug('Return code: %s' % self.__endless_command_return_code)
        if self.__endless_command_return_code!=0:
            return False

        return True

    def module_test_coverage(self, module_name, missing=False, as_json=False):
        """
        Display module coverage

        Args:
            module_name (string): module name
            missing (bool): display missing statements (default False)
            as_json (bool): return results as json

        Returns:
            string or dict according to as_json option
        """
        # checking module path
        path = os.path.join(config.MODULES_SRC, module_name, 'tests')
        if not os.path.exists(path):
            raise Exception('Specified module "%s" does not exist' % (module_name))

        module_version = self.__get_module_version(module_name)
        if module_version is None:
            raise Exception('Module version not found')

        coverage_file = self.__get_coverage_file(module_name, module_version)
        if not os.path.exists(coverage_file):
            raise Exception('No coverage file found. Did tests run ?')

        res = self.__coverage_simple_command(
            self.__get_module_tests_path(module_name),
            'report',
            '-m -i' if missing else '-i',
            coverage_file,
        )
        if res == False:
            raise Exception('Error executing coverage report')

        stdout = '\n'.join(res['stdout'])
        if not as_json:
            return stdout
        return self.__coverage_to_dict(stdout)

    def __get_core_tests_path(self):
        """
        Return core tests path

        Returns:
            string: core tests path
        """
        return '%s/tests' % (config.CORE_SRC)

    def __list_core_files(self, core_path):
        """
        Return all python files on core with their associated tests

        Args:
            core_path (string): core path

        Returns:
            dict: list of files with test file path::

                {
                    python file (string): test file (string)
                }

        """
        files = []
        tests_path = self.__get_core_tests_path()

        self.logger.info('Searching core files...')
        for root, _, filenames in os.walk(core_path):
            for filename in filenames:
                filepath = os.path.join(root, filename)

                # drop useless files
                if not filename.endswith('.py'):
                    self.logger.debug('  Drop file "%s": not python file' % filepath)
                    continue
                if filename.startswith('__init__'):
                    self.logger.debug('  Drop file "%s": python init file' % filepath)
                    continue
                if filename.startswith('test_'):
                    self.logger.debug('  Drop file "%s": test file' % filepath)
                    continue
                if root.endswith('libs/tests'):
                    self.logger.debug('  Drop file "%s": test lib' % filepath)
                    continue

                # build paths
                core_relative_path = root.replace('%s' % core_path, '')
                core_relative_path = core_relative_path[1:] if core_relative_path.startswith('/') else core_relative_path
                test_filepath = os.path.join(tests_path, core_relative_path, 'test_%s' % filename)
                self.logger.debug('  Found "%s" [%s]' % (filepath, test_filepath))
                files.append((filepath, test_filepath))

        return files

    def core_tests(self, display_coverage=False, display_test_output=False):
        """
        Execute core unit tests and display process output on stdout

        Args:
            display_coverage (bool): display coverage report (default False)
            display_test_output (bool): display unit test output (default True)

        Returns:
            bool: True if process succeed.
        """
        start = int(time.time())

        # clear previous results
        if self.__coverage_simple_command(self.__get_core_tests_path(), 'erase') == False:
            self.logger.error('Unable to clear previous tests results')
            return False

        # get files paths
        core_path = config.CORE_SRC
        files = self.__list_core_files(core_path)

        # execute tests
        self.logger.info('Running unit tests...')
        files_on_error = []
        files_on_success = []
        console_callback = self.__console_callback if display_test_output else self.__quiet_console_callback
        for filepath, test_filepath in files:
            self.logger.info('')
            self.logger.info('Testing %s using %s...' % (filepath.replace(core_path+'/', ''), test_filepath.replace(core_path+'/', '')))
            cmd = """
cd "%(core_tests_path)s"
coverage run --omit="*/lib/python*/*","*test_*.py" --concurrency=thread --parallel-mode %(test_file)s
            """ % {
                'core_tests_path': self.__get_core_tests_path(),
                'test_file': test_filepath,
            }
            self.logger.trace('Test cmd: %s' % cmd)
            self.__endless_command_running = True
            self.__reset_stds()
            c = EndlessConsole(cmd, console_callback, self.__console_end_callback)
            c.start()

            single_start = int(time.time())
            while self.__endless_command_running:
                time.sleep(0.25)

            single_duration = str(datetime.timedelta(seconds=(int(time.time()) - single_start)))
            self.logger.info('Duration %s' % single_duration)

            if self.__endless_command_return_code!=0:
                self.logger.debug('Command output:\n%s' % '\n'.join(self.__stderr))
                errors, failures, exception, notest = self.__get_coverage_errors_and_failures(test_filepath)
                files_on_error.append({
                    'filepath': filepath,
                    'errors': errors,
                    'failures': failures,
                    'exception': exception,
                    'notest': notest,
                })
            else:
                files_on_success.append({
                    'filepath': filepath
                })

        # coverage
        if display_coverage:
            coverage = self.core_test_coverage()
            logging.info(coverage)

        # display tests report
        duration = str(datetime.timedelta(seconds=(int(time.time()) - start)))
        self.logger.info('-' * 50)
        self.logger.info('Tests report:')
        self.logger.info('  Duration: %s' % duration)
        self.logger.info('  %d files succeed' % len(files_on_success))
        self.logger.info('  %d files on error' % len(files_on_error))
        if len(files_on_error) != 0:
            for file_on_error in files_on_error:
                self.logger.info('    - %(filepath)s: %(notest)s%(exception)s%(errors)s errors, %(failures)s failures' % file_on_error)
        self.logger.info('-' * 50)

        return True if len(files_on_error) == 0 else False

    def core_test_coverage(self, as_json=False):
        """
        Get core test coverage last results

        Args:
            as_json (bool): return output as json

        Returns:
            string or dict according to as_json option
        """
        # combine results
        if self.__coverage_simple_command(self.__get_core_tests_path(), 'combine', timeout=120.0) == False:
            raise Exception('Error preparing coverage results')

        # combine results
        res = self.__coverage_simple_command(self.__get_core_tests_path(), 'report', '-i', timeout=120.0)
        if res == False:
            raise Exception('Error generating coverage results')

        stdout = '\n'.join(res['stdout'])
        if not as_json:
            return stdout
        return self.__coverage-to_dict(stdout)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import subprocess
import time
from threading import Timer, Thread
try: # pragma: no cover
    from Queue import Queue, Empty
except ImportError: # pragma: no cover
    from queue import Queue, Empty  # python 3.x
import os
import signal
import logging
import re

class EndlessConsole(Thread):
    """
    Helper class to execute long command line (system update...)
    This kind of console doesn't kill command line after timeout. It just let command running
    until end of it or if user explicitely requests to stop (or kill) it.

    This class implements thread and is non blocking unless you use join function after starting it::

        c = EndlessConsole(mycmd, myclb, myendclb)
        c.join()

    Notes:
        Subprocess output async reading copied from https://stackoverflow.com/a/4896288
    """

    def __init__(self, command, callback, callback_end=None):
        """
        Constructor

        Args:
            command (string): command to execute
            callback (function): callback when message is received (the function will be called with 2
                                 arguments: stdout (string) and stderr (string))
            callback_end (function): callback when process is terminated (the function will be called
                                     with 2 arguments: return code (string) and killed (bool))
        """
        Thread.__init__(self, daemon=True, name='endlessconsole-%s' % getattr(callback, '__name__', 'unamed'))

        # members
        self.command = command
        self.callback = callback
        self.callback_end = callback_end
        self.logger = logging.getLogger(self.__class__.__name__)
        # self.logger.setLevel(logging.DEBUG)
        self.running = True
        self.killed = False
        self.__start_time = 0
        self.__stdout_queue = Queue()
        self.__stderr_queue = Queue()
        self.__stdout_thread = None
        self.__stderr_thread = None

    def __del__(self):
        """
        Destructor
        """
        self.__stop()

    def __enqueue_output(self, output, queue):
        for line in iter(output.readline, b''):
            if not self.running:
                break
            queue.put(line.decode('utf-8').rstrip())
        try:
            output.close()
        except Exception: # pragma: no cover
            pass

    def get_start_time(self):
        """
        Return process start time

        Returns:
            float: start timestamp (with milliseconds)
        """
        return self.__start_time

    def __stop(self):
        """
        Stop command line execution
        """
        self.running = False

    def stop(self):
        """
        Kill alias
        """
        self.kill()

    def kill(self):
        """
        Stop command line execution
        """
        self.logger.debug('Process killed manually')
        self.killed = True
        self.__stop()

    def __send_stds(self):
        """
        Read queues and send outputs if available

        Returns:
            True if something sent, False otherwise
        """
        if not self.callback:
            return False

        stdout = None
        stderr = None

        try:
            stdout = self.__stdout_queue.get_nowait()
        except Empty:
            pass
        except Exception: # pragma: no cover
            self.logger.exception('Error getting stdout queue')

        try:
            stderr = self.__stderr_queue.get_nowait()
        except Empty:
            pass
        except Exception: # pragma: no cover
            self.logger.exception('Error getting stderr queue')

        if stdout is not None or stderr is not None:
            try:
                self.callback(stdout, stderr)
            except Exception:
                self.logger.exception('Exception occured during EndlessCommand callback:')
            return True

        return False

    def run(self):
        """
        Console process
        """
        # launch command
        return_code = None
        self.__start_time = time.time()
        proc = subprocess.Popen(
            self.command,
            shell=True,
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            preexec_fn=os.setsid
        )
        pid = proc.pid
        self.logger.trace('PID=%d' % pid)

        if self.callback:
            # async stdout reading
            self.__stdout_thread = Thread(
                target=self.__enqueue_output,
                args=(proc.stdout, self.__stdout_queue),
                daemon=True,
                name='endlessconsole-stdout'
            )
            self.__stdout_thread.start()

            # async stderr reading
            self.__stderr_thread = Thread(
                target=self.__enqueue_output,
                args=(proc.stderr, self.__stderr_queue),
                daemon=True,
                name='endlessconsole-stderr'
            )
            self.__stderr_thread.start()

        # wait for end of command line
        while self.running:
            # check process status
            proc.poll()

            # read outputs and trigger callback
            self.__send_stds()

            # check end of command
            if proc.returncode is not None:
                return_code = proc.returncode
                self.logger.debug('Process is terminated with return code %s' % proc.returncode)
                break

            # pause
            time.sleep(0.25)

        # purge queues
        self.logger.trace('Purging outputs...')
        count = 0
        while self.__send_stds() or count <= 5:
            self.logger.trace(' purging...')
            count += 1
            time.sleep(0.05)
        self.logger.trace('Purge completed')

        # make sure all stds are closed
        try:
            proc.stdout.close()
        except Exception: # pragma: no cover
            pass
        try:
            proc.stderr.close()
        except Exception: # pragma: no cover
            pass

        # make sure process (and child processes) is really killed
        if self.killed and pid != 1:
            try:
                if ON_POSIX:
                    os.killpg(os.getpgid(pid), signal.SIGKILL)
                else: # pragma: no cover
                    proc.kill()
            except Exception: # pragma: no cover
                pass

        # process is over
        self.running = False

        # stop callback
        if self.callback_end:
            self.logger.trace('Call end callback')
            try:
                self.callback_end(return_code, self.killed)
            except Exception:
                self.logger.exception('Exception occured during EndlessCommand end callback:')


class Console():
    """
    Helper class to execute command lines.
    You can execute command right now using command method or after a certain amount of time using command_delayed
    """
    def __init__(self):
        """
        Constructor
        """
        # members
        self.timer = None
        self.__callback = None
        self.encoding = sys.getfilesystemencoding()
        self.last_return_code = None
        self.logger = logging.getLogger(self.__class__.__name__)
        # self.logger.setLevel(logging.DEBUG)

    def __del__(self):
        """
        Destroy console object
        """
        if self.timer: # pragma: no cover
            self.timer.cancel()

    def __process_lines(self, lines):
        """
        Remove end of line char for given lines and convert lines to unicode

        Args:
            lines (list): list of lines

        Results:
            list: input list of lines with eol removed
        """
        return [line.decode('utf-8').rstrip() for line in lines]

    def get_last_return_code(self):
        """
        DEPRECATED: use returncode from command result
        Return last executed command return code

        Returns:
            int: return code (can be None)
        """
        return self.last_return_code

    def command(self, command, timeout=2.0):
        """
        Execute specified command line with auto kill after timeout

        Notes:
            This function is blocking

        Args:
            command (string): command to execute
            timeout (float): wait timeout before killing process and return command result

        Returns:
            dict: result of command::

                {
                    returncode (int): command return code
                    error (bool): True if error occured,
                    killed (bool): True if command was killed,
                    stdout (list): command line output
                    stderr (list): command line error
                }

        """
        self.logger.trace('Launch command "%s"' % command)
        result = {
            'returncode': 0,
            'error': False,
            'killed': False,
            'stdout': [],
            'stderr': []
        }
        # check params
        if timeout is None or timeout <= 0.0:
            raise Exception('Timeout is mandatory and must be greater than 0')

        # launch command
        proc = subprocess.Popen(
            command,
            shell=True,
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            preexec_fn=os.setsid
        )
        pid = proc.pid

        # wait for end of command line
        done = False
        start = time.time()
        killed = False
        return_code = None
        while not done:
            # read stds to avoid deadlocks
            (stdout, stderr) = proc.communicate()
            if stdout:
                result['stdout'] += self.__process_lines(stdout.split(b'\n'))
            if stderr:
                result['stderr'] += self.__process_lines(stderr.split(b'\n'))

            # check if command has finished
            proc.poll()
            if proc.returncode is not None:
                # command terminated
                self.logger.trace('Command terminated with returncode %s' % proc.returncode)
                return_code = proc.returncode
                self.last_return_code = return_code
                done = True
                break

            # check timeout
            if time.time() > (start + timeout):
                # timeout is over, kill command
                pgid = os.getpgid(pid)
                self.logger.debug('Timeout over, kill command for PID=%s PGID=%s' % (pid, pgid))
                try:
                    if ON_POSIX:
                        os.killpg(pgid, signal.SIGKILL)
                    else: # pragma: no cover
                        proc.kill()
                except Exception: # pragma: no cover
                    pass
                killed = True
                break

            # pause
            time.sleep(0.125)

        # prepare result
        result['returncode'] = return_code
        if not killed:
            result['error'] = len(result['stderr']) > 0
        self.logger.trace('Result: %s' % result)

        # make sure all stds are closed
        try:
            proc.stdout.close()
        except Exception: # pragma: no cover
            pass
        try:
            proc.stderr.close()
        except Exception: # pragma: no cover
            pass

        # trigger callback (used for delayed command)
        if self.__callback:
            self.__callback(result)

        return result

    def command_delayed(self, command, delay, timeout=2.0, callback=None):
        """
        Execute specified command line after specified delay

        Args:
            command (string): command to execute
            delay (int): time to wait before executing command (milliseconds)
            timeout (float): timeout before killing command
            callback (function): function called when command is over. Callback will received command
                                 result as single function parameter

        Notes:
            See command function to have more details
        """
        self.__callback = callback
        self.timer = Timer(delay, self.command, [command, timeout])
        self.timer.daemon = True
        self.timer.name = 'commanddelayed'
        self.timer.start()


class AdvancedConsole(Console):
    """
    Create console with advanced feature like find function to match pattern on stdout
    """
    def __init__(self):
        """
        Constructor
        """
        Console.__init__(self)

    def find(self, command, pattern, options=re.UNICODE | re.MULTILINE, timeout=2.0, check_return_code=True, return_code=0):
        """
        Find all pattern matches in command stdout. Found order is preserved

        Args:
            command (string): command to execute
            pattern (string): search pattern
            options (flag): regexp flags (see https://docs.python.org/3/library/re.html#module-contents)
            timeout (float): timeout before killing command
            check_return_code (bool): check return code if True
            return_code (number): return code to check to detect failure

        Returns:
            list: list of matches::

                [
                    (group (string), matches in group (tuple)),
                    ...
                ]

        """
        # execute command
        res = self.command(command, timeout)
        if check_return_code and res['returncode'] != return_code:
            # command failed
            return []

        # parse command output
        content = '\n'.join(res['stdout'])
        return self.find_in_string(content, pattern, options)

    def find_in_string(self, string, pattern, options=re.UNICODE | re.MULTILINE):
        """
        Find all pattern matches in specified string. Found order is respected.

        Args:
            string (string): string to search in
            pattern (string): search pattern
            options (flag): regexp flags (see https://docs.python.org/2/library/re.html#module-contents)

        Returns:
            list: list of matches::

                [
                    (group (string), subgroups (tuple)),
                    ...
                ]

        """
        results = []
        matches = re.finditer(pattern, string, options)

        for _, match in enumerate(matches):
            group = match.group().strip()
            if len(group) > 0 and len(match.groups()) > 0:
                results.append((group, match.groups()))

        return results




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
from .console import AdvancedConsole
import logging
from . import config
from . import tools as Tools
import importlib
try:
    from cleep.common import CATEGORIES
    APP_CATEGORIES_CHECK_DISABLED = False
except:
    APP_CATEGORIES_CHECK_DISABLED = True
    logging.warn('Cleep module is not installed. Cleep application CATEGORIES validation is disabled')
import re
import inspect
import glob
import copy
import json
import time

class Check():
    """
    Handle module check:
        - check backend: variables, function
        - check frontend: config, desc.json, modImgSrc directive
        - run pylint on backend
    """

    PYLINTRC = """[MASTER]
extension-pkg-whitelist=
ignore=git
ignore-patterns=
jobs=1
load-plugins=
persistent=yes
suggestion-mode=yes
unsafe-load-any-extension=no

[MESSAGES CONTROL]
confidence=
disable=print-statement,
        parameter-unpacking,
        unpacking-in-except,
        old-raise-syntax,
        backtick,
        long-suffix,
        old-ne-operator,
        old-octal-literal,
        import-star-module-level,
        non-ascii-bytes-literal,
        invalid-unicode-literal,
        raw-checker-failed,
        bad-inline-option,
        locally-disabled,
        locally-enabled,
        file-ignored,
        suppressed-message,
        useless-suppression,
        deprecated-pragma,
        apply-builtin,
        basestring-builtin,
        buffer-builtin,
        cmp-builtin,
        coerce-builtin,
        execfile-builtin,
        file-builtin,
        long-builtin,
        raw_input-builtin,
        reduce-builtin,
        standarderror-builtin,
        unicode-builtin,
        xrange-builtin,
        coerce-method,
        delslice-method,
        getslice-method,
        setslice-method,
        no-absolute-import,
        old-division,
        dict-iter-method,
        dict-view-method,
        next-method-called,
        metaclass-assignment,
        indexing-exception,
        raising-string,
        reload-builtin,
        oct-method,
        hex-method,
        nonzero-method,
        cmp-method,
        input-builtin,
        round-builtin,
        intern-builtin,
        unichr-builtin,
        map-builtin-not-iterating,
        zip-builtin-not-iterating,
        range-builtin-not-iterating,
        filter-builtin-not-iterating,
        using-cmp-argument,
        eq-without-hash,
        div-method,
        idiv-method,
        rdiv-method,
        exception-message-attribute,
        invalid-str-codec,
        sys-max-int,
        # bad-python3-import,
        # deprecated-string-function,
        deprecated-str-translate-call,
        # deprecated-itertools-function,
        # deprecated-types-field,
        next-method-defined,
        # dict-items-not-iterating,
        # dict-keys-not-iterating,
        # dict-values-not-iterating,
        # deprecated-operator-function,
        # deprecated-urllib-function,
        xreadlines-attribute,
        # deprecated-sys-function,
        exception-escape,
        comprehension-escape,
        too-few-public-methods,
        too-many-instance-attributes,
        trailing-newlines,
        len-as-condition,
        logging-not-lazy,
        broad-except,
        missing-module-docstring,
        relative-beyond-top-level, # to remove when pylint bug resolved
enable=c-extension-no-member

[REPORTS]
evaluation=10.0 - ((float(5 * error + warning + refactor + convention) / statement) * 10)
output-format=text
reports=no
score=yes

[REFACTORING]
max-nested-blocks=5
never-returning-functions=optparse.Values,sys.exit

[TYPECHECK]
contextmanager-decorators=contextlib.contextmanager
# zmq.{LINGER,REQ,ROUTER,NOBLOCK} are dynamically generated and so pylint
# doesn't see them, causing false positives.
generated-members=PAIR,RCVHWM,SNDHWM,SNDTIMEO,RCVTIMEO,netifaces.*
ignore-mixin-members=yes
ignore-on-opaque-inference=yes
ignored-classes=optparse.Values,thread._local,_thread._local
ignored-modules=
missing-member-hint=yes
missing-member-hint-distance=1
missing-member-max-choices=1

[VARIABLES]
additional-builtins=
allow-global-unused-variables=yes
callbacks=cb_,
          _cb
dummy-variables-rgx=_+$|(_[a-zA-Z0-9_]*[a-zA-Z0-9]+?$)|dummy|^ignored_|^unused_
ignored-argument-names=_.*|^ignored_|^unused_
init-import=no
redefining-builtins-modules=six.moves,past.builtins,future.builtins,io,builtins

[SPELLING]
max-spelling-suggestions=4
spelling-dict=
spelling-ignore-words=
spelling-private-dict-file=
spelling-store-unknown-words=no

[FORMAT]
expected-line-ending-format=
ignore-long-lines=^\s*(# )?<?https?://\S+>?$
indent-after-paren=4
indent-string='    '
max-line-length=130
max-module-lines=1000
no-space-check=trailing-comma,
               dict-separator
single-line-class-stmt=no
single-line-if-stmt=no

[BASIC]
argument-naming-style=snake_case
attr-naming-style=snake_case
bad-names=foo,
          bar,
          baz,
          toto,
          tutu,
          tata
class-attribute-naming-style=any
class-naming-style=PascalCase
const-naming-style=UPPER_CASE
docstring-min-length=-1
function-naming-style=snake_case
good-names=i,
           j,
           k,
           ex,
           Run,
           to,
           _
include-naming-hint=no
inlinevar-naming-style=any
method-naming-style=snake_case
module-naming-style=snake_case
name-group=
no-docstring-rgx=^_
property-classes=abc.abstractproperty
variable-naming-style=snake_case

[MISCELLANEOUS]
notes=FIXME,
      XXX,
      TODO

[SIMILARITIES]
ignore-comments=yes
ignore-docstrings=yes
ignore-imports=no
min-similarity-lines=4

[LOGGING]
logging-modules=logging

[IMPORTS]
allow-wildcard-with-all=no
analyse-fallback-blocks=no
deprecated-modules=regsub,
                   TERMIOS,
                   Bastion,
                   rexec
ext-import-graph=
import-graph=
int-import-graph=
known-standard-library=
known-third-party=enchant

[DESIGN]
max-args=5
max-attributes=7
max-bool-expr=5
max-branches=12
max-locals=15
max-parents=7
max-public-methods=20
max-returns=6
max-statements=50
min-public-methods=2

[CLASSES]
defining-attr-methods=__init__,
                      __new__,
                      setUp
exclude-protected=_asdict,
                  _fields,
                  _replace,
                  _source,
                  _make
valid-classmethod-first-arg=cls
valid-metaclass-classmethod-first-arg=mcs

[EXCEPTIONS]
overgeneral-exceptions=Exception

    """

    VERSION_UNRELEASED = 'UNRELEASED'

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def check_backend(self, module_name, module_author=None):
        """
        Check backend

        Args:
            module_name (string): module name to check
            module_author (string): if specified check if author is corresponding

        Returns:
            dict: backend informations::

            {
                errors (list): list of errors
                warnings (list): list of warnings
                metadata (dict): module metadata
                files (dict): collection of files::
                    {
                        module (dict): main module informations
                        events (list): events informations
                        formatters (list): formatters informations
                        drivers (list): drivers informations
                        misc (list): others files informations
                    }
            }

        """
        if not os.path.exists(os.path.join(config.MODULES_DST, module_name)):
            raise Exception('Module "%s" does not exist' % module_name)

        # get module instance
        try:
            module_ = importlib.import_module(u'cleep.modules.%s' % (module_name))
            app_filename = getattr(module_, 'APP_FILENAME', module_name)
            del module_
            module_ = importlib.import_module('cleep.modules.%s.%s' % (module_name, app_filename))
            class_name = next((item for item in dir(module_) if item.lower() == app_filename.lower()), None)
            class_ = getattr(module_, class_name or '', None)
        except Exception as e:
            self.logger.exception('Unable to load application "%s". Please check your code' % module_name)
            raise Exception('Unable to load application "%s". Please check your code' % module_name) from e
        if not class_:
            raise Exception('Main class was not found for app "%s". Application class must have the same name than app name' % module_name)

        # build metadata and list of files
        metadata = self.__build_metadata(class_, module_author)
        files = self.__build_files_list(class_)

        return {
            'errors': metadata['errors'] + files['errors'],
            'warnings': metadata['warnings'] + files['warnings'],
            'metadata': metadata['metadata'],
            'files': {
                'module': files['module'],
                'events': files['events'],
                'drivers': files['drivers'],
                'formatters': files['formatters'],
                'misc': files['misc'],
            }
        }
        
    def __build_files_list(self, class_):
        """
        Build module list of files

        Args:
            class_ (Class): loaded module class

        Returns:
            dict: list of modules files::

            {
                errors (list): list of errors
                warnings (list): list of warnings
                module (dict): file infos for main module
                events (list): list of events infos
                formatters (list): list or formatters infos
                drivers (list): list of drivers infos
                misc (list): list of miscellaneous files
            }
                
        """
        errors = []
        warnings = []

        # get all application files
        all_files = self.__get_backend_files(class_)
        self.logger.debug('All files: %s' % all_files)

        # analyse files
        events = self.__get_files_for_kind(all_files['files'], {'endswith': 'event', 'class': 'Event'})
        errors += events['errors']
        warnings += events['warnings']
        drivers = self.__get_files_for_kind(all_files['files'], {'endswith': 'driver', 'class': 'Driver'})
        errors += drivers['errors']
        warnings += drivers['warnings']
        formatters = self.__get_files_for_kind(all_files['files'], {'endswith': 'formatter', 'class': 'ProfileFormatter'})
        errors += formatters['errors']
        warnings += formatters['warnings']

        # fill misc files
        found_files = ([e['fullpath'] for e in events['files']] + 
                       [f['fullpath'] for f in formatters['files']] +
                       [d['fullpath'] for d in drivers['files']])
        self.logger.debug('Found files: %s' % found_files)
        misc = [f for f in all_files['files'] if f['fullpath'] not in found_files]
        misc += all_files['initpy']
        self.logger.debug('Misc: %s' % misc)

        # check __init__.py
        initpy_fullpaths = [a_file['fullpath'] for a_file in all_files['initpy']]
        if not self.__has_init_py(initpy_fullpaths, all_files['folders']):
            errors.append('Some __init__.py files are missing in root folder or sub folders')

        return {
            'errors': errors,
            'warnings': warnings,
            'module': all_files['module'],
            'misc': sorted(misc, key=lambda k: k['fullpath']),
            'events': sorted(events['files'], key=lambda k: k['fullpath']),
            'formatters': sorted(formatters['files'], key=lambda k: k['fullpath']),
            'drivers': sorted(drivers['files'], key=lambda k: k['fullpath']),
        }

    def __has_init_py(self, initpy, folders):
        """
        Check if __init__.py files exists in all folders

        Args:
            initpy (list): list of existing __init__.py fullpaths
            folders (list): list of application folders

        Returns:
            bool: True if all __init__.py exists
        """
        return all([os.path.join(folder, '__init__.py') in initpy for folder in folders])

    def __get_files_for_kind(self, all_files, kind):
        """
        Return files infos according to specified kind

        Args:
            all_files (list): list of files in module folder
            kind (dict): kind of files to search for::
            
                {
                    class (string): must be Event|ProfileFormatter|Driver
                    endswith (string): event|formatter|driver
                }

        Returns:
            list: list of files of requested kind
        """
        out = {
            'errors': [],
            'warnings': [],
            'files': []
        }

        for a_file in all_files:
            # drop useless files
            if not a_file['filename'].lower().endswith(kind['endswith'] + '.py'):
                continue

            # load file
            try:
                mod_name = a_file['filename'].replace('.py', '')
                parts = Tools.full_split_path(a_file['fullpath'])
                mod_ = importlib.import_module('cleep.modules.%s.%s' % (parts[-2], mod_name))
            except Exception as e:
                self.logger.exception('Error loading file "%s"' % a_file['fullpath'])
                out['errors'].append('Error loading file "%s". Please check file [%s]' % (
                    a_file['fullpath'], str(e)
                ))
                continue

            # check class name
            class_name = next((item for item in dir(mod_) if item.lower() == mod_name.lower()), None)
            if not class_name:
                out['errors'].append('Error loading file "%s": class name should have the same name than filename' % a_file['fullpath'])
                continue
            class_ = getattr(mod_, class_name)

            # check base class
            if not any([True for c in inspect.getmro(class_) if c.__name__ == kind['class']]):
                out['errors'].append('Error loading file "%s": class "%s" should inherit from "%s" due to its name. Please fix it' % (
                    a_file['fullpath'],
                    class_name,
                    kind['class'],
                ))
                continue

            out['files'].append({
                'fullpath': a_file['fullpath'],
                'filename': a_file['filename'],
                'path': a_file['path'],
                'classname': class_name,
            })

        return out

    def __get_backend_files(self, class_):
        """
        Get list of backend files

        Args:
            class_ (Class): loadde module class

        Returns:
            list: list of found files::

            {
                initpy (list): list of __init__.py fullpaths
                module (dict): main module infos::
                    {
                        fullpath (string): fullpath
                        filename (string): filename
                        path (string): path within module
                    },
                files (dict): other files infos::
                    [
                        {
                            fullpath (string): fullpath
                            filename (string): filename
                            path (string): path within module
                        },
                        ...
                    ],
                folders (list): list of module subfolders
            }

        """
        class_path = inspect.getfile(class_).replace('.pyc', '.py')
        module_path = class_path.rsplit('/',1)[0]

        fullpaths = glob.glob(module_path + '/**/*', recursive=True)
        out = {
            'module': {},
            'files': [],
            'initpy': [],
            'folders': [],
        }
        for fullpath in fullpaths:
            # drop useless files
            fileext = os.path.splitext(fullpath)[1]
            if '__pycache__' in fullpath:
                continue
            if any([True for f in ['.pylintrc', ] if fullpath.find(f) >= 0]):
                continue
            if fileext != '.py':
                continue

            if fullpath.endswith('__init__.py'):
                # handle __init__.py file
                out['initpy'].append({
                    'fullpath': fullpath,
                    'filename': os.path.split(fullpath)[1],
                    'path': fullpath.split('modules/')[1],
                })
            elif fullpath == class_path:
                # handle main module
                out['module'] = {
                    'fullpath': fullpath,
                    'filename': os.path.split(fullpath)[1],
                    'path': fullpath.split('modules/')[1],
                }
            else:
                # handle other file
                out['files'].append({
                    'fullpath': fullpath,
                    'filename': os.path.split(fullpath)[1],
                    'path': fullpath.split('modules/')[1],
                })

            # add scanned folders
            folder = os.path.dirname(fullpath)
            if folder not in out['folders']:
                out['folders'].append(folder)

        return out

    def __build_metadata(self, class_, module_author=None):
        """
        Build module metadata from module constants

        Args:
            class_ (Class): loaded module class
            module_author (string): module author to check

        Returns:
            dict: metadata::

            {
                metadata (dict): list of module properties
                errors (list): list of errors
                warnings (list): list of warnings
            }

        """
        check = self.__check_backend_constants(class_, module_author)

        return {
            'metadata': {
                'author': getattr(class_, 'MODULE_AUTHOR', None),
                'description': getattr(class_, 'MODULE_DESCRIPTION', None),
                'longdescription': getattr(class_, 'MODULE_LONGDESCRIPTION', None),
                'category': getattr(class_, 'MODULE_CATEGORY', None),
                'deps': getattr(class_, 'MODULE_DEPS', []),
                'version': getattr(class_, 'MODULE_VERSION', None),
                'tags': getattr(class_, 'MODULE_TAGS', []),
                'country': getattr(class_, 'MODULE_COUNTRY', None),
                'urls': {
                    'info': getattr(class_, 'MODULE_URLINFO', None),
                    'help': getattr(class_, 'MODULE_URLHELP', None),
                    'site': getattr(class_, 'MODULE_URLSITE', None),
                    'bugs': getattr(class_, 'MODULE_URLBUGS', None),
                },
                'price': getattr(class_, 'MODULE_PRICE', None),
                'label': getattr(class_, 'MODULE_LABEL', class_.__name__),
            },
            'errors': check['errors'],
            'warnings': check['warnings'],
        }

    def __check_constant(self, constant):
        """
        Check specified constant

        Args:
            parameters (list): list of parameters to check::

                {
                    name (string): parameter name
                    type (type): parameter primitive type (str, bool...)
                    none (bool): True if parameter can be None
                    empty (bool): True if string value can be empty
                    value (any): parameter value
                    validator (function): validator function. Take value in parameter and must return bool
                    message (string): custom message to return instead of generic error
                },
                ...

        Returns:
            string: message in case of error or warning, None if nothing to report

        """
        # check None
        if ('none' not in constant or ('none' in constant and not constant['none'])) and constant['value'] is None:
            return 'Constant "%s" is missing' % constant['name']

        # check value
        if constant['value'] is None:
            # nothing else to check, constant value is allowed as None above
            return None

        # check type
        if not isinstance(constant['value'], constant['type']):
            return 'Constant "%s" has wrong type ("%s" instead of "%s")' % (
                constant['name'],
                type(constant['value']).__name__,
                constant['type'].__name__,
            )

        # use validator if provided
        if 'validator' in constant and not constant['validator'](constant['value']):
            return constant['message'] if 'message' in constant else 'Constant "%s" is invalid (specified="%s")' % (
                constant['name'],
                constant['value'],
            )

        # check empty
        if (('empty' not in constant or ('empty' in constant and not constant['empty'])) and
                getattr(constant['value'], '__len__', None) and
                len(constant['value']) == 0):
            return constant['message'] if 'message' in constant else 'Constant "%s" is empty (specified="%s")' % (
                constant['name'],
                constant['value'],
            )

        return None

    def __check_backend_constants(self, class_, module_author=None):
        """
        Check module constants

        Args:
            class_ (Class): loaded module class
            module_author (string): module author to check

        Returns:
            dict: errors and warnings::

            {
                errors (list): list of errors
                warnings (list): list of warnings
            }

        """
        out = {
            'errors': [],
            'warnings': [],
        }

        # MODULE_AUTHOR
        author = getattr(class_, 'MODULE_AUTHOR', None)
        msg = self.__check_constant({'name': 'MODULE_AUTHOR', 'type': str, 'value': author})
        if msg:
            out['errors'].append(msg)
        if module_author and author and module_author.lower() != author.lower():
            out['errors'].append('Application author must be the same than repository: %s != %s' % (author, module_author))

        # MODULE_DESCRIPTION
        msg = self.__check_constant({'name': 'MODULE_DESCRIPTION', 'type': str, 'value': getattr(class_, 'MODULE_DESCRIPTION', None)})
        if msg:
            out['errors'].append(msg)

        # MODULE_LONGDESCRIPTION
        msg = self.__check_constant({'name': 'MODULE_LONGDESCRIPTION', 'type': str, 'value': getattr(class_, 'MODULE_LONGDESCRIPTION', None)})
        if msg:
            out['errors'].append(msg)

        # MODULE_CATEGORY
        if not APP_CATEGORIES_CHECK_DISABLED:
            msg = self.__check_constant({
                'name': 'MODULE_CATEGORY',
                'type': str,
                'value': getattr(class_, 'MODULE_CATEGORY', None),
                'validator': lambda val: val in CATEGORIES.ALL,
                'message': 'MODULE_CATEGORY must be filled with existing categories. See cleep.common.CATEGORIES'
            })
            if msg:
                out['errors'].append(msg)

        # MODULE_DEPS
        msg = self.__check_constant({'name': 'MODULE_DEPS', 'type': list, 'value': getattr(class_, 'MODULE_DEPS', None), 'empty': True})
        if msg:
            out['errors'].append(msg)

        # MODULE_VERSION
        msg = self.__check_constant({
            'name': 'MODULE_VERSION',
            'type': str,
            'value': getattr(class_, 'MODULE_VERSION', None),
            'validator': lambda val: re.compile(r'\d+\.\d+\.\d+').match(val),
            'message': 'MODULE_VERSION must follow semver rules https://semver.org/',
        })
        if msg:
            out['errors'].append(msg)

        # MODULE_TAGS
        msg = self.__check_constant({'name': 'MODULE_TAGS', 'type': list, 'value': getattr(class_, 'MODULE_TAGS', None)})
        if msg:
            out['warnings'].append(msg)

        # MODULE_URLINFO
        msg = self.__check_constant({'name': 'MODULE_URLINFO', 'type': str, 'value': getattr(class_, 'MODULE_URLINFO', None)})
        if msg:
            out['warnings'].append(msg)

        # MODULE_URLHELP
        msg = self.__check_constant({'name': 'MODULE_URLHELP', 'type': str, 'value': getattr(class_, 'MODULE_URLHELP', None)})
        if msg:
            out['warnings'].append(msg)

        # MODULE_URLSITE
        msg = self.__check_constant({'name': 'MODULE_URLSITE', 'type': str, 'value': getattr(class_, 'MODULE_URLSITE', None)})
        if msg:
            out['warnings'].append(msg)

        # MODULE_URLBUGS
        msg = self.__check_constant({'name': 'MODULE_URLBUGS', 'type': str, 'value': getattr(class_, 'MODULE_URLBUGS', None)})
        if msg:
            out['warnings'].append(msg)

        # MODULE_COUNTRY
        msg = self.__check_constant({
            'name': 'MODULE_URLCOUNTRY',
            'type': str,
            'value': getattr(class_, 'MODULE_COUNTRY', None),
            'none': True,
            'validator': lambda val: len(val) == 2,
            'message': 'Constant MODULE_COUNTRY must be ISO3166-2 compatible code https://fr.wikipedia.org/wiki/ISO_3166-2',
        })
        if msg:
            out['errors'].append(msg)

        # MODULE_PRICE
        msg = self.__check_constant({
            'name': 'MODULE_PRICE',
            'type': float,
            'value': getattr(class_, 'MODULE_PRICE', None),
            'none': True,
        })
        if msg:
            out['errors'].append(msg)

        # MODULE_LABEL
        msg = self.__check_constant({
            'name': 'MODULE_LABEL',
            'type': str,
            'value': getattr(class_, 'MODULE_LABEL', None),
            'none': True
        })
        if msg:
            out['errors'].append(msg)

        return out

    def check_frontend(self, module_name):
        """
        Check frontend

        Args:
            module_name (string): module name to check

        Returns:
            dict: frontend informations::
            
            {
                errors (list): list of errors
                warnings (list): list of warnings
                icon (string): icon name
                files (list): files informations::
                    [
                        {
                            fullpath (string): file fullpath
                            filename (string): filename
                            path (string): module path
                            usage (string): file usage
                            extension (string): file extension
                        }
                        ...
                    ]
            }

        """
        if not os.path.exists(os.path.join(config.MODULES_DST, module_name)):
            raise Exception('Module "%s" does not exist' % module_name)

        out = {
            'errors': [],
            'warnings': [],
            'files': [],
            'icon' : None
        }

        # get all files
        all_files = self.__get_frontend_files(module_name)
        
        # check desc.json
        desc_json_info = self.__get_desc_json(all_files)
        self.logger.debug('desc_json_info: %s' % desc_json_info)
        if not desc_json_info:
            out['errors'].append('desc.json file is missing. Please add it following Cleep recommandation')
        else:
            desc_json = self.__check_desc_json(desc_json_info)
            self.logger.debug('desc_json: %s' % desc_json)
            out['icon'] = desc_json['content'].get('icon', None)
            out['errors'] += desc_json['errors']
            out['warnings'] += desc_json['warnings']
        if not desc_json['content']:
            raise Exception('Invalid desc.json file. Please check content.')

        # give files usage according to desc.json
        if not out['errors']:
            all_checks = self.__check_frontend_files(module_name, all_files, desc_json)
            out['errors'] += all_checks['errors']
            out['warnings'] += all_checks['warnings']
            out['files'] = all_checks['files']

        return out

    def __check_modimgsrc_directive(self, module_name, js_files, all_files):
        """
        Check if developer uses mod-img-src directive to display its images

        Args:
            module_name (string): module name
            js_files (list): list of js files
            all_files (dict): list of all files

        Returns:
            tuple: list of warnings and list of found images::

            (
                ['warning1', 'warning2', ...],
                ['mod/img1.png', 'mod/img2.jpg', ...]
            )

        """
        # init
        cacheds = []
        warnings = []
        founds = []

        # get images
        image_files = [a_file for a_file in all_files['files'] if a_file['extension'] in ('.jpg', '.jpeg', '.gif', '.png', '.webp')]
        html_files = [a_file for a_file in all_files['files'] if a_file['extension'] == '.html']
        logging.debug('Image files %s' % image_files)
        logging.debug('Html files %s' % html_files)

        # no image, no need to go further
        if len(image_files) == 0:
            return warnings, founds

        # cache html files content
        for html_file in html_files:
            with open(html_file['fullpath'], 'r') as fdesc:
                cacheds.append('\n'.join(fdesc.readlines()))

        # check directive usage for found images
        for image_file in image_files:
            pattern = r"mod-img-src\s*=\s*[\"']\s*%s\s*[\"']" % image_file['path'].replace(module_name+'/', '')
            self.logger.debug('Mod-img-src pattern: %s' % pattern)
            found = False
            for cached in cacheds:
                matches = re.finditer(pattern, cached, re.MULTILINE)
                if len(list(matches)) > 0:
                    logging.debug('Image "%s" found' % image_file['path'])
                    found = True
                    founds.append(image_file['path'])
            if not found:
                warnings.append('Image "%s" may not be displayed properly because mod-img-src directive wasn\'t used' % image_file['filename'])

        return warnings, founds

    def __check_frontend_files(self, module_name, all_files, desc_json):
        """
        Check frontend files

        Args:
            module_name (string): module name
            all_files (list): list of all frontend files
            desc_json (dict): desc.json analyze result

        Returns:
            dict: check results::

            {
                errors (list): list of errors
                warnings (list): list of warnings
                files (list): list of all files updated with usage info
            }

        """
        out = {
            'errors': [],
            'warnings': [],
            'files': copy.deepcopy(all_files['files']),
        }
        content = desc_json['content']
        config_files = []
        global_files = []
        res_files = []
        pages_files = []
        html_files = []
        js_files = []
        css_files = []

        # fill found files
        if 'global' in content:
            global_files += content['global']['js'] if 'js' in content['global'] else []
            js_files += content['global']['js'] if 'js' in content['global'] else []
            global_files += content['global']['html'] if 'html' in content['global'] else []
            html_files += content['global']['html'] if 'html' in content['global'] else []
            global_files += content['global']['css'] if 'css' in content['global'] else []
            css_files += content['global']['css'] if 'css' in content['global'] else []
        if 'config' in content:
            config_files += content['config']['js'] if 'js' in content['config'] else []
            js_files += content['config']['js'] if 'js' in content['config'] else []
            config_files += content['config']['html'] if 'html' in content['config'] else []
            html_files += content['config']['html'] if 'html' in content['config'] else []
            config_files += content['config']['css'] if 'css' in content['config'] else []
            css_files += content['config']['css'] if 'css' in content['config'] else []
        if 'pages' in content:
            for page in content['pages']:
                pages_files += content['pages'][page]['js'] if 'js' in content['pages'][page] else []
                js_files += content['pages'][page]['js'] if 'js' in content['pages'][page] else []
                pages_files += content['pages'][page]['html'] if 'html' in content['pages'][page] else []
                html_files += content['pages'][page]['html'] if 'html' in content['pages'][page] else []
                pages_files += content['pages'][page]['css'] if 'css' in content['pages'][page] else []
                css_files += content['pages'][page]['css'] if 'css' in content['pages'][page] else []
        if 'res' in content:
            res_files += content['res']

        # check files place
        for a_file in js_files:
            if os.path.splitext(a_file)[1] != '.js':
                out['warnings'].append('File "%s" should not be in "js" section. Please fix it' % a_file)
        for a_file in html_files:
            if os.path.splitext(a_file)[1] not in ('.html', '.htm'):
                out['warnings'].append('File "%s" should not be in "html" section. Please fix it' % a_file)
        for a_file in css_files:
            if os.path.splitext(a_file)[1] != '.css':
                out['warnings'].append('File "%s" should not be in "css" section. Please fix it' % a_file)
        for a_file in res_files:
            if os.path.splitext(a_file)[1] not in ('.png', '.jpg', '.jpeg', '.gif', '.webp'):
                out['warnings'].append('File "%s" seems not to have a supported image format. Please convert it' % a_file)

        # check images
        warnings, found_images = self.__check_modimgsrc_directive(module_name, js_files, all_files)
        out['warnings'] += warnings

        # give flags to files
        for a_file in out['files']:
            if a_file['filename'] in global_files:
                a_file['usage'] = 'GLOBAL'
            elif a_file['filename'] in config_files:
                a_file['usage'] = 'CONFIG'
            elif a_file['filename'] in pages_files:
                a_file['usage'] = 'PAGES'
            elif a_file['filename'] in res_files:
                a_file['usage'] = 'RES'
            elif a_file['filename'] == 'desc.json':
                a_file['usage'] = 'CORE'
            elif a_file['path'] in found_images:
                a_file['usage'] = 'CONFIG'
            else:
                out['warnings'].append('File "%s" is unused' % a_file['path'])
                a_file['usage'] = 'UNUSED'

        # search for missing files
        paths = [item['path'].replace(module_name + '/', '') for item in all_files['files']]
        self.logger.debug('paths: %s' % paths)
        for a_file in global_files:
            a_file not in paths and out['errors'].append('File "%s" specified in desc.json "global" section is missing' % a_file)
        for a_file in config_files:
            a_file not in paths and out['errors'].append('File "%s" specified in desc.json "config" section is missing' % a_file)
        for a_file in res_files:
            a_file not in paths and out['errors'].append('File "%s" specified in desc.json "res" section is missing' % a_file)
        for a_file in pages_files:
            a_file not in paths and out['errors'].append('File "%s" specified in desc.json "pages" section is missing' % a_file)

        return out

    def __check_desc_json(self, desc_json_info):
        """
        Check desc.json content

        Args:
            desc_json_info: desc.json file info

        Returns:
            dict: check results::

            {
                errors (list): list of errors
                warnings (list): list of warnings
            }

        """
        out = {
            'errors': [],
            'warnings': [],
            'content': None,
        }

        try:
            with open(desc_json_info['fullpath']) as json_file:
                content = json.load(json_file)
                out['content'] = content
        except:
            if self.logger.getEffectiveLevel() == logging.DEBUG:
                self.logger.exception('Error loading desc.json:')
            out['errors'].append('Error loading "%s". Please check file content.' % desc_json['fullpath'])
            return out

        # check content
        if not all([key in list(content.keys()) for key in ('config', 'global', 'icon')]):
            out['errors'].append('Invalid desc.json content. At least one mandatory key is missing')

        # check global section
        if 'global' in content:
            if 'js' in content['global'] and not isinstance(content['global']['js'], list):
                out['errors'].append('Invalid global.js section: it must be an array')
            if 'html' in content['global'] and not isinstance(content['global']['html'], list):
                out['errors'].append('Invalid global.html section: it must be an array')
            if 'css' in content['global'] and not isinstance(content['global']['css'], list):
                out['errors'].append('Invalid global.css section: it must be an array')

        # check config section
        if 'config' in content:
            if 'js' in content['config'] and not isinstance(content['config']['js'], list):
                out['errors'].append('Invalid config.js section: it must be an array')
            if 'html' in content['config'] and not isinstance(content['config']['html'], list):
                out['errors'].append('Invalid config.html section: it must be an array')
            if 'css' in content['config'] and not isinstance(content['config']['css'], list):
                out['errors'].append('Invalid config.css section: it must be an array')

        # check res section
        if 'res' in content and not isinstance(content['res'], list):
            out['errors'].append('Invalid res section: it must be an array')

        # check pages section
        if 'pages' in content:
            for page in content['pages']:
                if 'js' in content['pages'][page] and not isinstance(content['pages'][page]['js'], list):
                    out['errors'].append('Invalid pages.%s.js section: it must be an array' % page)
                if 'html' in content['pages'][page] and not isinstance(content['pages'][page]['html'], list):
                    out['errors'].append('Invalid pages.%s.html section: it must be an array' % page)
                if 'css' in content['pages'][page] and not isinstance(content['pages'][page]['css'], list):
                    out['errors'].append('Invalid pages.%s.css section: it must be an array' % page)

        return out

    def __get_desc_json(self, all_files):
        """
        Get desc.json file

        Args:
            all_files (list): list of frontend files

        Returns:
            dict: desc.json file infos. None if file not found
        """
        self.logger.debug('all_files: %s' % all_files['files'])
        return next((a_file for a_file in all_files['files'] if a_file['filename'] == 'desc.json'), None)

    def __get_frontend_files(self, module_name):
        """
        Get list of frontend files

        Args:
            module_name (string): module name

        Returns:
            list: list of found files::

            {
                initpy (list): list of __init__.py fullpaths
                module (dict): main module infos::
                    {
                        fullpath (string): fullpath
                        filename (string): filename
                        path (string): path within module
                    },
                files (dict): other files infos::
                    [
                        {
                            fullpath (string): fullpath
                            filename (string): filename
                            path (string): path within module
                        },
                        ...
                    ],
                folders (list): list of module subfolders
            }

        """
        module_path = os.path.join(config.MODULES_HTML_DST, module_name)
        fullpaths = glob.glob(module_path + '/**/*', recursive=True)
        out = {
            'files': [],
        }
        for fullpath in fullpaths:
            # drop some files
            filename = os.path.split(fullpath)[1]
            if filename.startswith('.') or filename.startswith('~') or filename.endswith('.tmp'):
                continue

            # drop directories
            if os.path.isdir(fullpath):
                continue

            # store file infos
            out['files'].append({
                'fullpath': fullpath,
                'filename': filename,
                'path': fullpath.split('modules/')[1],
                'extension': os.path.splitext(fullpath)[1],
            })

        return out

    def check_scripts(self, module_name):
        """
        Check scripts files

        Args:
            module_name (string): module name

        Returns:
            dict: scripts infos::

            {
                errors (list): list of errors
                warnings (list): list of warnings
                files (list): list of files::
                    [
                        {
                            fullpath (string): file fullpath
                            filename (string): filename
                            path (string): module path
                        }
                        ...
                    ]
            }

        """
        if not os.path.exists(os.path.join(config.MODULES_DST, module_name)):
            raise Exception('Module "%s" does not exist' % module_name)

        scripts_path = os.path.join(config.MODULES_SRC, module_name, 'scripts')
        fullpaths = glob.glob(scripts_path + '/**/*', recursive=True)
        out = {
            'errors': [],
            'warnings': [],
            'files': [],
        }
        for fullpath in fullpaths:
            # store file infos
            out['files'].append({
                'fullpath': fullpath,
                'filename': os.path.split(fullpath)[1],
                'path': fullpath.split('modules/%s/' % module_name)[1],
            })

        return out

    def check_tests(self, module_name):
        """
        Check test files

        Args:
            module_name (string): module name

        Returns:
            dict: scripts infos::

            {
                errors (list): list of errors
                warnings (list): list of warnings
                files (list): list of files::
                    [
                        {
                            fullpath (string): file fullpath
                            filename (string): filename
                            path (string): module path
                        }
                        ...
                    ]
            }

        """
        if not os.path.exists(os.path.join(config.MODULES_DST, module_name)):
            raise Exception('Module "%s" does not exist' % module_name)

        scripts_path = os.path.join(config.MODULES_SRC, module_name, 'tests')
        fullpaths = glob.glob(scripts_path + '/**/*', recursive=True)
        out = {
            'errors': [],
            'warnings': [],
            'files': [],
        }
        for fullpath in fullpaths:
            # drop some files
            filename = os.path.split(fullpath)[1]
            ext = os.path.splitext(fullpath)[1],
            if filename.startswith('.') or filename.startswith('~') or filename.endswith('.tmp'):
                continue
            if '__pycache__' in fullpath:
                continue

            # store file infos
            out['files'].append({
                'fullpath': fullpath,
                'filename': filename,
                'path': fullpath.split('modules/%s/' % module_name)[1],
            })

        # check mandatory files
        filenames = [item['filename'] for item in out['files']]
        if '__init__.py' not in filenames:
            out['errors'].append('Mandatory "__init__.py" file is missing in tests folder. Please add it')
        if 'test_%s.py' % module_name not in filenames:
            out['errors'].append('Tests file "test_%s.py" must exists in tests folder' % module_name)

        return out

    def check_code_quality(self, module_name, rewrite_pylintrc=False):
        """
        Check code quality running pylint on backend. Add .pylintrc if file is missing

        Args:
            module_name (string): module name
            rewrite_pylintrc (bool): True to rewrite .pylintrc file

        Returns:
            dict: result of pylint::

            {
                errors (list): list of errors::
                {
                    filename (string): {
                        code (string): pylint code
                        msg (string): pylint message
                    }
                }
                warnings (list): list of warnings::
                {
                    filename (string): {
                        code (string): pylint code
                        msg (string): pylint message
                    }
                }
                score (float): code quality score (/10)
            }

        """
        if not os.path.exists(os.path.join(config.MODULES_DST, module_name)):
            raise Exception('Module "%s" does not exist' % module_name)

        backend_path = os.path.join(config.MODULES_SRC, module_name, 'backend')
        pylintrc_path = os.path.join(backend_path, '.pylintrc')
        if not os.path.exists(pylintrc_path) or rewrite_pylintrc:
            self.logger.debug('Create default "%s" file' % pylintrc_path)
            with open(pylintrc_path, 'w') as pylintrc_file:
                pylintrc_file.write(self.PYLINTRC)
            time.sleep(3.0)

        # launch pylint
        self.logger.debug('Launch pylint')
        cmd = 'cd "%s"; pylint *.py' % backend_path
        pattern = r'^(.*?\.py):(\d+):(\d+): ([CRWEF]\d+): (.*)$|^.*at (\d+\.\d+)\/10.*$'
        console = AdvancedConsole()
        results = console.find(cmd, pattern, timeout=60, check_return_code=False)
        self.logger.debug('Results: %s' % results)

        out = {
            'errors': [],
            'warnings': [],
            'score': 0.0,
        }
        if len(results) == 0:
            out['errors'].append({
                'code': None,
                'msg': 'Internal error: linter execution failed. Please check your code.',
            })
        for result in results:
            group = list(filter(None, result[1]))
            if len(group) == 5:
                # pylint output
                msg = '%s [%s %s:%s]' % (group[4], group[0], group[1], group[2])
                if group[3].startswith('E') or group[3].startswith('F'):
                    out['errors'].append({
                        'code': group[3],
                        'msg': msg
                    })
                else:
                    out['warnings'].append({
                        'code': group[3],
                        'msg': msg,
                    })
            elif len(group) == 1:
                # score
                out['score'] = float(group[0])

        self.logger.debug('Code quality output: %s' % out)
        return out

    def check_changelog(self, module_name):
        """
        Check module changelog

        Args:
            module_name (string): module name

        Returns:
            dict: changelog metadata::

            {
                version (string): module version
                changelog (string): latest version changelog
                unreleased (bool): True if version unreleased
            }

        """
        if not os.path.exists(os.path.join(config.MODULES_DST, module_name)):
            raise Exception('Module "%s" does not exist' % module_name)

        # search for changelog.md file
        changelog_path = None
        fullpaths = glob.glob(os.path.join(config.MODULES_SRC, module_name) + '/*')
        for fullpath in fullpaths:
            if os.path.basename(fullpath).lower() == 'changelog.md':
                changelog_path = fullpath
        if not changelog_path:
            raise Exception('Application changelog "changelog.md" does not exist. Please create it following https://keepachangelog.com/en/1.0.0/')
        self.logger.debug('Using changelog "%s"' % changelog_path)

        with open(changelog_path) as changelog_file:
            lines = changelog_file.readlines()

        # read all sections
        sections = []
        current_section = None
        for line in lines:
            if line.startswith('## '):
                current_section = [line,]
                sections.append(current_section)
            elif current_section:
                current_section.append(line)

        # get first section and search for version
        first_section = sections[0] if len(sections) > 0 else []
        found_version = None
        unreleased = False
        if len(first_section) > 0:
            pattern = re.compile(r'\d+\.\d+\.\d+')
            match = pattern.search(first_section[0])
            found_version = match.group() if match else None

            pattern = re.compile(r'%s' % self.VERSION_UNRELEASED)
            match = pattern.search(first_section[0].split('\n')[0])
            unreleased = match is not None
        self.logger.debug('Found version: %s' % found_version)
        self.logger.debug('Found unreleased: %s' % unreleased)

        return {
            'version': found_version,
            'changelog': ''.join(first_section).strip(),
            'unreleased': unreleased,
        }




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
from .console import Console
import logging
from . import config

class File():
    """
    Handle file operations
    """

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def core_sync(self):
        """
        Synchronize core content between source and execution folders
        """
        # check vars
        if not config.CORE_DST:
            raise Exception('Cleep needs to be installed before using this command')

        c = Console()
        cmd = """
/bin/mkdir -p "%(HTML_DST)s/"
/usr/bin/rsync -av "%(REPO_DIR)s/cleep/" "%(CORE_DST)s/" --exclude "/tests/" --exclude "modules/**/" --exclude "*__pycache__*" --delete --exclude "*.pyc"
/usr/bin/rsync -av "%(REPO_DIR)s/html/" "%(HTML_DST)s/" --delete --exclude "js/modules/" --exclude "*node_modules*"
/usr/bin/rsync -av "%(REPO_DIR)s/bin/cleep" "%(BIN_DST)s/cleep"
/usr/bin/rsync -av "%(REPO_DIR)s/medias/sounds" "%(MEDIA_DST)s/sounds/" --delete
    """ % {'REPO_DIR': config.REPO_DIR, 'CORE_DST': config.CORE_DST, 'HTML_DST':config.HTML_DST, 'BIN_DST':config.BIN_DST, 'MEDIA_DST':config.MEDIA_DST}
        self.logger.debug('Coresync cmd: %s' % cmd)
        resp = c.command(cmd, 15)
        self.logger.debug('Coresync stdout:\n%s' % '\n'.join(resp['stdout']))
        if resp['stderr']:
            self.logger.debug('Coresync stderr:\n%s' % '\n'.join(resp['stderr']))
        if resp['error'] or resp['killed']:
            self.logger.error('Error occured while sync core content: %s' % ('killed' if resp['killed'] else resp['stderr']))
            return False

        return True

    def module_sync(self, module_name):
        """ 
        Copy specified module content to valid execution location
        """
        if not os.path.exists(os.path.join(config.MODULES_SRC, module_name)):
            sys.exit('Module "%s" doesn\'t exist [%s]' % (module_name, os.path.join(config.MODULES_SRC, module_name)))
    
        if not os.path.exists(os.path.join(config.MODULES_DST, module_name)):
            os.makedirs(os.path.join(config.MODULES_DST, module_name))

        mod_backend_src = os.path.join(config.MODULES_SRC, module_name, 'backend/')
        mod_backend_dst = os.path.join(config.MODULES_DST, module_name)
        mod_frontend_src = os.path.join(config.MODULES_SRC, module_name, 'frontend/')
        mod_frontend_dst = os.path.join(config.MODULES_HTML_DST, module_name)
        mod_scripts_src = os.path.join(config.MODULES_SRC, module_name, 'scripts/')
        mod_scripts_dst = os.path.join(config.MODULES_SCRIPTS_DST, module_name)

        c = Console()
        cmd = """ 
if [ -d "%(BACKEND_SRC)s" ]; then
    /bin/mkdir -p "%(BACKEND_DST)s"
    rsync -av "%(BACKEND_SRC)s" "%(BACKEND_DST)s" --delete --exclude "*.pyc" --exclude "*__pycache__*"
fi
if [ -d "%(FRONTEND_SRC)s" ]; then
    /bin/mkdir -p "%(FRONTEND_DST)s"
    rsync -av "%(FRONTEND_SRC)s" "%(FRONTEND_DST)s" --delete --exclude "*node_modules*"
fi
if [ -d "%(SCRIPTS_SRC)s" ]; then
    /bin/mkdir -p "%(SCRIPTS_DST)s"
    rsync -av "%(SCRIPTS_SRC)s" "%(SCRIPTS_DST)s" --delete --exclude "*__pycache__*"
fi
    """ % { 
            'BACKEND_SRC': mod_backend_src, 'BACKEND_DST': mod_backend_dst,
            'FRONTEND_SRC': mod_frontend_src, 'FRONTEND_DST': mod_frontend_dst,
            'SCRIPTS_SRC': mod_scripts_src, 'SCRIPTS_DST': mod_scripts_dst
        }
        self.logger.debug('Modsync cmd: %s' % cmd)
        resp = c.command(cmd, 15) 

        self.logger.debug('Modsync stdout:\n%s' % '\n'.join(resp['stdout']))
        if resp['stderr']:
            self.logger.debug('Modsync stderr:\n%s' % '\n'.join(resp['stderr']))
        if resp['error'] or resp[u'killed']:
            self.logger.error(u'Error occured while sync module content: %s' % (u'killed' if resp[u'killed'] else resp[u'stderr']))
            return False

        return True

